Unit FuncUnit;

{ Copyright (C) 1998-2004 Idan Nof (GNU GPL) }
{ Read the attached files "README.TXT" and/or "GPL.TXT" for more details }

{ Last Updated: 02/07/2004 }
{********************************} Interface {*******************************}

{ Set compiler directives and user symbols }
{$I SYMBOLS.INC }

Uses MathUnit; { Complex numbers and polynomials }

{ Set floating point type }
{$I FLOAT.INC  }

Const
  MaxFuncs =  9;  { Maximum number of functions at one time}
  MaxDegree=100;  { Largest degree polynomial }

Var
  TF: Byte; {Total Functions }

Type
  DegreeType = 0..MaxDegree;      { Possible polynomial degrees }

  { Linked-List }
  PolyPtr = ^Polynom;             { A forward declaration to a polynom. }
  Polynom = Record                { polynom. Record                     }
              Link : PolyPtr;     { Link to next coeff                  }
              Coeff: Float;       { Coefficient                         }
            End;

  { Linked-List }
  RootPtr  = ^EachRoot;        { A forward declaration to EachRoot }
  EachRoot = Record            { EachRoot Record                   }
               Link : RootPtr; { Link to next root                 }
               Root : Complex; { Root itself                       }
             End;

  PolyType = Record                 { PolyType Record     }
               Polynom: PolyPtr   ; { Pointer to polynom. }
               Exp    : DegreeType; { Degree of polynom.  }
               Roots    : RootPtr ; { Roots Stack         }
               RootCount: Byte    ; { Number of Roots     }
               NoRealRoots:Boolean ;{ No Real Roots ?     }
             End;

  RatioType = Record                { Rational func. Record          }
                Num,Denom: PolyType;{ Num. & Denom. of type PolyType }
              End;

  TypeOfFunc = (Pol,Rat,Exp,Log{,General,GenRat}); { Set of known functions }
  FuncPtr = ^FuncType;                    { A pointer to a general function }

  { Linked-List }
  MinfoPTR = ^MInfoList;         { Forward declaration to additional info }
  MInfoList = Record             { Additional info record                 }
                Link: MInfoPTR;  { Link to next item                      }
                Math: Float;     { Item itself                            }
              End;

  FuncType = Record                      { General function record     }
               TypeOf : TypeOfFunc;      { Type of function            }
          {*}  BInfo : LongInt ;         { Boolean packed info         }
          {**} MInfo : MInfoPTR;         { Additional info             }
               MaxX,MaxY: Float;         { 'Interesting' range         }
               Case TypeOfFunc of        { Split structure declaration }
                     Pol : (Poly , PolyDiff     : PolyType ); { Polynomial  }
                     Rat : (Ratio, RatioDiff    : RatioType); { Rational    }
                     Exp : (BaseE, FreeE, RootE : Float );    { Exponential }
                     Log : (BaseL, FreeL, RootL : Float );    { Logarithmic }
{                  General : (Num,Denom           : String  ); { General     }
               End;

  FuncArr = Array[1..MaxFuncs] of FuncPtr;

Var
 Funcs: FuncArr;            { Array of MaxFuncs pointers to function }


  {*}
  { BInfo is of Longint size and represents 32 distinct boolean             }
  { values (bits 0..31)                                                     }
  {                                                                         }
  { * Not all of this is yet implemented and may be changed.                }
  {                                                                         }
  { Bit Number / (True value) states that the function                      }
  { -------------------------------------------------                       }
  { 0          - Is defined for every (Real) X                              }
  { 1          - Is defined when X=0                                        }
  { 2          - Has got an horizontal asymptote                            }
  { 3          - Has got a curved asymptote                                 }
  { 4..7 => n  - Number of distinct derivative zero's (and undefined values)}
  { 8..8+n     - Gives a postive values (+) on the matching interval.       }

  {**}
  { MInfo is a pointer to a linked list and it's order is :                }
  { Item Number /           Item Description                               }
  { -------------------------------------------------------                }
  { 1           - Intersection point with the Y-Axis (assuming one exists) }
  { 2           - Horizontal Asymptote (y=b)         (   "      "     "  ) }
  { 3           - Curved Asym. slope (m)             (   "      "     "  ) }
  { 4           - Curved Asym. free coeff (n)        (   "      "     "  )
  {                                                                        }
  { * Curved asymptote is one of the form y=mx+n                           }


Procedure Init;                           { Init   }
Procedure Quit;                           { Finish }

Procedure InitFunc(Var Func: FuncPtr);    { Allocate Memory        }
Procedure ChooseType(Func: FuncPtr);      { Choosing function type }
Procedure Get(Var Func: FuncPtr);         { Reading the func.      }
Procedure Analyze(Func: FuncPtr);         { Analyzing the func.    }
Procedure WriteAnalyze(Var Func: FuncPtr);{ Write analyze report   }
Procedure Sketch(Funcs: FuncArr);         { Sketch Drawing         }
Procedure DisposeOf(Var Funcs: FuncArr);  { Dispose Memory         }
Function  AddAnotherFunc: Boolean;        { Add Function ?         }
Function  DoYouWantAnother: Boolean;      { Another Function ?     }

{$IFDEF DEBUG}
Procedure CheckF(Func: FuncPtr);          { Check validity    (DEBUG) }
{$ENDIF}

{****************************} Implementation {******************************}

Uses Crt,Graph,
     LPT{Printer},MSMouse{Mouse},Jump{Error Handler},VESA{VESA Detection} ;

{Include Jordan's Video Modes Numbers }
{$I SVGA256.INC}

Type
  bitNum = 0..31;     { Bit Num }
  bit    = 0..1;

  SignType = -1..1;

  HelpType   = Array[DegreeType] of Float;
  C_HelpType = Array[DegreeType] of Complex;

  PolyHelpPtr = ^PolyHelp;
  PolyHelp = Record
               Coeff: Float;
               Deg  : Byte;
               Link : PolyHelpPtr;
             End;

  OptionsRec = Record
     ErrorHandler, Complex: Boolean;
     Mouse, RealFormat, LPTPort : Byte;
     PrintDPI: Word;
     GraphFP, GDriver, GMode: Integer;
  End;

  Pixels16x16 = Array[0..15,0..15] of Byte;

Const
  Copyright = 'The Function Analyzer  Version 1.02j  Copyright (C) 1998-2004 Idan Nof';
  Esc  = #27;
  Bell = #7;         { Bell's ASCII Code }
  Options : OptionsRec = (ErrorHandler: True; Complex: True; Mouse: 0;
                          RealFormat: 3; LPTPort: 1;
                          PrintDPI: 100; GraphFP: -1;
                          GDriver : 0; GMode: 0);

  NewInit     : Boolean = True;
  Skip_Message: Boolean = False;
  Invalid_Operation_Flag: Boolean = False;

 BWCross: Array[1..32] of Word =
         ($FFFF,  { 1111 1111 1111 1111 }
          $FFFF,  { 1111 1111 1111 1111 }
          $FFFF,  { 1111 1111 1111 1111 }
          $FFFF,  { 1111 1111 1111 1111 }
{Screen}  $FFFF,  { 1111 1111 1111 1111 }
{ Mask }  $FFFF,  { 1111 1111 1111 1111 }
          $FFFF,  { 1111 1111 1111 1111 }
          $FF7F,  { 1111 1111 0111 1111 }
          $FFFF,  { 1111 1111 1111 1111 }
          $FFFF,  { 1111 1111 1111 1111 }
          $FFFF,  { 1111 1111 1111 1111 }
          $FFFF,  { 1111 1111 1111 1111 }
          $FFFF,  { 1111 1111 1111 1111 }
          $FFFF,  { 1111 1111 1111 1111 }
          $FFFF,  { 1111 1111 1111 1111 }
          $FFFF,  { 1111 1111 1111 1111 }

          $0080,  { 0000 0000 1000 0000 }
          $0080,  { 0000 0000 1000 0000 }
          $0080,  { 0000 0000 1000 0000 }
          $0080,  { 0000 0000 1000 0000 }
{Cursor}  $0080,  { 0000 0000 1000 0000 }
{ Mask }  $0080,  { 0000 0000 1000 0000 }
          $0080,  { 0000 0000 1000 0000 }
          $7F7F,  { 0111 1111 0111 1111 }
          $0080,  { 0000 0000 1000 0000 }
          $0080,  { 0000 0000 1000 0000 }
          $0080,  { 0000 0000 1000 0000 }
          $0080,  { 0000 0000 1000 0000 }
          $0080,  { 0000 0000 1000 0000 }
          $0080,  { 0000 0000 1000 0000 }
          $0080,  { 0000 0000 1000 0000 }
          $0000); { 0000 0000 0000 0000 }

Var
   MouseOn : Boolean;
   OldExit : Pointer;
   MyAddr  : JumpRecord;
   SVGA256 : Integer;

   mx,my: Word;
   HotX,HotY: Integer;
   OldPixels: Pixels16x16;
   ColorCross: Array[1..2] of Pixels16x16;

   Epsilon: Float;

   {$IFDEF DEBUG}
   FMem: LongInt;
   {$ENDIF      }


Procedure MyExit; Far;
{This error-handler is activated only in graphics mode }
{It get called ("automatically") when a Run-Time error is generated }

Var Key: Char;
Begin
   {Restore the old exit proc so if error happens the program wouldn't }
   {go into infinite loop                                              }
   ExitProc:=OldExit;

   { ErrorAddr=Nil in case of normal termination }
   If ErrorAddr<>Nil Then Begin
      RestoreCRTMode;
      TextBackground(LightBlue);
      TextColor(LightGray);
      ClrScr;
      Write('Run-Time Error: ');
      case ExitCode of
           200 : Writeln('Division by zero');
           205 : Writeln('Floating point overflow');
           206 : Writeln('Floating point underflow');
           207 : Writeln('Invalid floating point operation');
           else  Writeln(ExitCode,' (Unexptected)'); { = Unexpected Error }
      end;
      Write('<ESC> to quit, any other key to re-sketch using defaults');

      While KeyPressed do ReadKey;

      Key:=ReadKey;
      Writeln;
      If Key=#27{ESC} then
         Begin
           {$IFDEF DEBUG}
           DisposeOf(Funcs);
           {$ENDIF}

           RunError(ExitCode)   { Move on the ExitCode to the OS }
         End
      else
         Begin
            ErrorAddr:=Nil;
            LongJump(MyAddr); { Jump to saved address }
         End;
   end;
End;

Function Is_Invalid_Operation: Boolean;
{ Returns the status of the last arithmetic operation which modified
  the "Invalid_Operation_Flag" , since calling this function;

  Returns TRUE if invalid operation performed since reading the
  "Invalid_Operation_Flag" via the "Is_Invalid_Operation" function }

Var Temp: Boolean;
Begin
   Temp := Invalid_Operation_Flag;
   Invalid_Operation_Flag := False;
   Is_Invalid_Operation := Temp;
End;

{----------------------------------------------------------------------------}
{$IFDEF DEBUG}
Procedure CheckF(Func: FuncPtr);
{ For debugging purposes : Checks polynomials and rationals to see if that
  their degree and number of roots is not larger than the maximum declared.
  (what is likely to happen if the program has got a bug in one of the
  related, pointer based, routines) }

Const
    Str_TypeOfFunc : Array[TypeOfFunc] of String[3] =
                     ('Pol','Rat','Exp','Log');

Var i: Integer;
    OK: Boolean;

Begin
   OK:=True;
   For i:=1 to TF do
       With Func^ do
            Case Typeof Of
            Pol:
               Begin
                  With Poly do
                       OK := (Exp<=MaxDegree) and (RootCount<=MaxDegree);
                  With PolyDiff do
                       OK := (OK) and (Exp<=MaxDegree) and (RootCount<=MaxDegree);
               End;
            Rat:
               Begin
                  With Ratio do
                       Begin
                          With Num do
                               OK := (Exp<=MaxDegree) and (RootCount<=MaxDegree);
                          With Denom do
                               OK := (OK) and (Exp<=MaxDegree) and (RootCount<=MaxDegree);
                       End;
                  With RatioDiff do
                       Begin
                          With Num do
                               OK := (Exp<MaxDegree) and (RootCount<MaxDegree);
                          With Denom do
                               OK := (OK) and (Exp<MaxDegree) and (RootCount<MaxDegree);
                       End;
               End;
       End;
   If Not OK then
      Begin
         Writeln;
         Writeln('Error !');
         Writeln('Function #',i);
         Writeln('Func^.TypeOf = ',Str_TypeOfFunc[Func^.TypeOf]);
         Writeln('Press <ENTER> to continue.');
         Readln;
      End;
End; {CheckF}
{$ENDIF }

{----------------------------------------------------------------------------}
{ Link EGAVGA BGI Driver }
Procedure EGAVGA_Driver; external; {$L EGAVGA.OBJ }

{ Link SVGA256 BGI Driver }
procedure SVGA256_Driver; external;{$L SVGA256.OBJ }

{----------------------------------------------------------------------------}
Procedure WritelnF(S: String; Margin: Byte);
{ Prints the String S is with a right 'Margin' format }
Begin
   Writeln(' ':79-Length(S)-Margin,S);
End; {WritelnF}

{----------------------------------------------------------------------------}
Procedure WriteF(S: String; Margin: Byte);
{ Prints the String S is with a right 'Margin' format }
Begin
   Write(' ':79-Length(S)-Margin,S);
End; {WriteF}

{----------------------------------------------------------------------------}
Function GetBit(b : LongInt; n : bitNum) : bit;
{ Returns the value of the n-th bit in b }
begin
  GetBit := bit(odd(b shr n));
end; {GetBit}

{----------------------------------------------------------------------------}
Function SetBit( b : LongInt; n : bitNum) : LongInt;
{ Returns the binary number b while forcing the n-th bit to '1' }
begin
  setBit := b or (1 shl n);
end; {SetBit}

{----------------------------------------------------------------------------}
Function ClrBit(b : LongInt; n : bitNum) : LongInt;
{ Returns the binary number b while forcing the n-th bit to '0' }
begin
  clrBit := b and hi($FEFF shl n);
end; {ClrBit}

{----------------------------------------------------------------------------}
Function BigInt(A,B: Integer): Integer;
{ Returns the largest of two integers }
Begin
   If A>=B then
             BigInt:=A
           else
             BigInt:=B;
End;

{----------------------------------------------------------------------------}
Function PwrXY (b, n : Float) : Float;
{ b (B,n rational )}
{ System.Exp = Exp using system unit (since Exp is already defined)}
BEGIN
  IF (n = 0.0) THEN
    PwrXY := 1.0
  ELSE
    IF (b <= 0.0) AND (Frac(n) = 0.0) THEN
      IF (Frac(n / 2)) = 0.0 THEN
        PwrXY := System.Exp(n * Ln(Abs(b)))
      ELSE
        PwrXY := - System.Exp(n * Ln(Abs(b)))
    ELSE
      PwrXY := System.Exp(n * Ln(b));
END; {PwrXY}

{----------------------------------------------------------------------------}
Procedure ClearVec(Var Vec: HelpType);
{ Zero all cells in HelpType array }
Var i: Byte;
Begin
   For i:=0 to MaxDegree do
       Vec[i] := 0;
End; {ClearVec}

Procedure c_ClearVec(Var Vec: c_HelpType);
{ Zero all cells in C_HelpType array }
Var i: Byte;
Begin
   For i:=0 to MaxDegree do
       Vec[i] := CZero;
End; {c_ClearVec}

{----------------------------------------------------------------------------}
Function RealF(X: Float): String;
{1.44654365 = 1.446 (user precision + internal rounding) }
{1.0        = 1     {Cut all non-significant zeros       }
{1.54000000 = 1.54  (all significant digits up to user precision + rounding }
{1.0005     = 1     (Internal Rounding + Cut Zeros       }
Var T: String;
Begin
   Str(X:0:Options.RealFormat,T);
   While T[Length(T)]='0' do Delete(T,Length(T),1);
   If T[Length(T)]='.' then Delete(T,Length(T),1);

   RealF := T;
End;

{$IFDEF VERBOSE}
Function RealF_(X: Float): String;
Var T: String;
Begin
   If (FRAC(X)=0) or (X>100*EPSILON) then Str(X:0:3*Options.RealFormat,T)
                     else Str(X,T);

   While T[Length(T)]='0' do Delete(T,Length(T),1);
   If T[Length(T)]='.' then Delete(T,Length(T),1);

   RealF_ := T;
End;
{$ENDIF}

{----------------------------------------------------------------------------}
Procedure Graph_Abort(Message: String; ErrorCode: Integer);
{Return to textmode, print error message, error code and exit}
Begin
   TextMode(LastMode);
   Writeln('Fatal graphics error: '+Message);
   Writeln;
   Writeln('   Status: ',GraphErrorMsg(ErrorCode),'.');
   Halt(1);
End;

Procedure GraphMode;
Var ErrCode: Integer;
Begin
   { NewInit is true whenever we have used CloseGraph procedure
     or didn't had at least one successful InitGraph call
     (Whenever we need to call to InitGraph )
   }

   With Options do
       If NewInit then
          Begin
             InitGraph(GDriver, GMode,'');
             NewInit := False;
          End
          else SetGraphMode(GMode);

   ErrCode := GraphResult;
   If ErrCode <> grOk then Graph_Abort('Error Initializing Graphics Mode.',ErrCode);
End;

{----------------------------------------------------------------------------}
Function UpStr(St: String): String;
Var i: Integer;
Begin
   For i:=1 to Length(St) do
       St[i] := UpCase(St[i]);

   Upstr := St;
End;

{----------------------------------------------------------------------------}
Function Int2Str(Int: Longint): String;
Var St: String;
Begin
   Str(Int,St);
   Int2Str := St;
End;

{----------------------------------------------------------------------------}
Function PutBetween(X1,X2: Byte; St: String): String;
Var
  Center, i: Integer;
  TempStr: String;
Begin
   TempStr:='';
   Center:=X1 + (X2 - X1 - Length(St)) div 2;
   For i:=1 to Center do
       TempStr := TempStr + ' ';
   PutBetween := TempStr+St;
End;

{----------------------------------------------------------------------------}
Procedure Init;
{ Init: Parameters processing & Graphics detection }

Var Err: Integer;

Procedure Process_Parameters;
Var
  i,R: Integer;
  S, S2 : String;
Begin
   For i:=1 to ParamCount do With Options do
       Begin
          S := UpStr(ParamStr(i));
          If (S[1]='/') and (S[2] in ['F','G','L','M','P','R','V']) then
             Begin
                S2 := Copy(S,3,Length(S)-2);
                Val(S2,R,Err);
                If (Err=0) and (R>=0) then
                   Case S[2] of
                        { Force Mouse ON/OFF}
                        'F': If (R>=1) or (R<=2) then Mouse:=R;

                        'G': If (R>=0) and (R<=8) then GraphFP := R;

                        { Select LPT Port }
                        'L': If R<=3 then
                                Begin
                                   LPTPort := R;
                                   If LPTPort<>1 then SelectPrinter(LPTPort-1);
                                End;

                        { Force Video Mode }
                        'M': If (R>=0) and (R<=6) then GMode := R;

                        'P': If R<=17 then RealFormat := R;

                        { Select printer resolution }
                        'R': If (R<=600) and (R<>0) then PrintDPI := R;

                        { Force Video driver }
                        'V': If (R=1) or (R=2) then GDRiver:=R;
                   End; {Case}
             End
          else If (S[1]='/') and (S[2] in ['C','D']) then
                  Case S[2] of
                       'C': Complex:=False;
                       'D': ErrorHandler:=False;
                  End {Case}
          else
            Begin
               Writeln(Copyright);
               Writeln;
               Writeln('Syntax: fa [switch switch...]');
               Writeln;
               Writeln('Switch is one or more of the following: ');
               Writeln;
               Writeln('/H,/? - This help screen');
               Writeln('/C    - Disable output of polynomials complex roots');
               Writeln('/D    - Disable internal error handler');
               Writeln('/Fn   - Force mouse (default autodetect; ON=1, OFF=2)');
               Writeln('/Pn   - Set precision in general to n digits after point (default n=3; 0n17) ');
               Writeln('/Gn   - Force precision in graph to n digits (auto & dynamic by default; 0n8)');
               Writeln('/Rx   - Set printer resolution to n DPI (default x=100; 1x600)');
               Writeln('/Ln   - Set printer port to n (default LPT1; 1n3)');
               Writeln('/Vn   - Force Video Driver to n (default autodetect; VGA=1, SVGA=2)');
               Writeln('/Mn   - Force Video Mode to n (only affects if Driver is VGA or SVGA)');
               Halt;
            End;
       End;
End; {ProcessParameters}

Procedure InitGraphics;
Begin
   {Add the SVGA driver to the BGI table }
   SVGA256 := InstallUserDriver('SVGA256', @DetectVESA256);

   {Inform the graphics system of the linked EGAVGA BGI driver }
   If (RegisterBGIDriver(@EGAVGA_Driver) < 0) then
      Graph_Abort('Error registering EGAVGA driver.',GraphResult);

   With Options do
        Begin
           If GDriver=Detect then
              Begin
                 If DetectVESA256=grError then
                    Begin
                       DetectGraph(GDriver,GMode);
                       Err:=GraphResult;
                       If Err<>grOK then
                          Graph_Abort('Can''t detect graphics hardware.',Err);
                    End;
              End
           else If GDriver=1 then GDriver:=VGA
           else If GDriver=2 then GDriver:=0;

           If GDriver=0 then
              Begin
                 {Inform the graphics system of the linked SVGA BGI driver }
                 If (RegisterBGIDriver(@SVGA256_Driver) < 0) then
                    Graph_Abort('Error registering SVGA driver.',GraphResult);

                 GDriver:=SVGA256;
                 If GMode=0 then GMode := SVGA800x600x256;
              End
           else If (GDriver=VGA) and (GMode=0) then GMode:=VGAHi;

        End; { With }
End; {InitGraphics}

function calceps: float;
{calceps.pas ==
  This function returns the machine EPSILON or floating point tolerance,
  the smallest positive real number such that 1.0 + EPSILON > 1.0.
  EPSILON is needed to set various tolerances for different algorithms.
  While it could be entered as a constant, I prefer to calculate it, since
  users tend to move software between machines without paying attention to
  the computing environment. Note that more complete routines exist.
}
const MaxI = 100; { Set this so always : 1/(2^MaxI) < EPSILON }
var
  e,e0: float;
  i: integer;
begin {calculate machine epsilon}
  e0 := 1; i:=0;
  repeat
    e0 := e0/2; e := 1+e0;  i := i+1;
  until (e=1.0) or (i=MaxI); {note safety check}
  e0 := e0*2;
  calceps:=e0;
end; {calceps}

Begin {Init}
   Process_Parameters;

   {$IFDEF DEBUG}
   FMem:=MemAvail;
   {$ENDIF      }

   InitGraphics;
   EPSILON := CalcEps;
End; {Init}

{----------------------------------------------------------------------------}
Procedure ChooseType(Func: FuncPtr); { Choosing Function Type }
Var
  Key: Char;
Begin
   TextBackGround(Blue);
   TextColor(LightGray);
   ClrScr;
   Writeln(PutBetween(1,80,Copyright));
   Writeln;
   If TF=1 then
      Begin
         Writeln(PutBetween(1,80,'Welcome to ''The Function Analyzer'''));
         Writeln(PutBetween(1,80,'------------------------------------'));
      End
   else
      Begin
         Writeln('     Available memory: '+Int2Str(MemAvail)+' bytes.');
         Writeln;
      End;

   Writeln;
   Writeln;
   Writeln('   Please choose the function''s type:');
   Writeln;
   Writeln('        1. Polynomial.............................P(x)= An*x^++A1*x+A0');
   Writeln;
   Writeln('        2. Rational...............................R(x)= P1(x) / P2(x)');
   Writeln;
   Writeln('        3. Exponential............................E(x)= A^x+B');
   Writeln;
   Writeln('        4. Logarithmic............................L(x)= log x(A)+B');
   GotoXY(39,7);
   Repeat
       Key:=ReadKey;
   Until Key in ['1'..'4',#27];

   If Key=ESC then  { Quick exit key }
      Begin
         Quit;
         DisposeOf(Funcs);
         Halt;
      End;

   Func^.TypeOf:=TypeOfFunc( Ord(Key) - 49 );
   { Modify function type                           }
   { The structure TypeOfFunc is a zero-based array }
   { which means that, TypeOfFunc(0) = Pol          }

End; {ChooseFuncType}

{----------------------------------------------------------------------------}
Function EvalPolynom(Poly: PolyType; X: Float): Float;
{ Evaluate polynomial function value in X }
{ This one requires n additions & 2n multiplications for polynom. of degree n}

Var Index: PolyPtr;
    i: Integer;
    Temp, Mul: Float;
Begin
   Mul   := 1;
   Index := Poly.Polynom;
   Temp  := Index^.Coeff; {Free Coeff.}
   For i:=1 to Poly.Exp do
       Begin
          Index := Index^.Link;
          Mul := Mul * X;
          Temp := Temp + (Mul*Index^.Coeff);
       End;
   EvalPolynom:=Temp;
End; {EvalPolynom}

{----------------------------------------------------------------------------}
Procedure C_EvalPolynom(Poly: PolyType; Z: Complex; Var Result: Complex);
{ Evaluate polynomial function value in Z }
{ This one requires n additions & 2n multiplications for polynom. of degree n}

Var Index: PolyPtr;
    i: Integer;
    T, Temp, Mul: Complex;
Begin
   CMake(1,0,Mul);
   Index := Poly.Polynom;
   CMake(Index^.Coeff,0,Temp ); {Free Coeff.}
   For i:=1 to Poly.Exp do
       Begin
          Index := Index^.Link;
          CMult(Mul,Z,Mul);

          CMake(Index^.Coeff,0,T);
          CMult(Mul,T,T);
          CAdd( Temp, T, Temp );
       End;
   Result:=Temp;
End; {c_EvalPolynom}

{----------------------------------------------------------------------------}
Function EvalExp(Func: FuncPtr; X: Float): Float;
{ Evaluate exponential function value in X }
Var T: Float;
Begin
   With Func^ do
        If BaseE>0 then T := PwrXY(BaseE,X) + FreeE
                   else Invalid_Operation_Flag := True;
   EvalExp := T;
End; {EvalExp}

{----------------------------------------------------------------------------}
Function EvalLog(Func: FuncPtr; X: Float): Float;
{ Evaluate logarithmic function value in X }
Var T: Float;
Begin
   With Func^ do
        If X>0 then T := ( ln(X)/ln(BaseL) ) + FreeL
               else Invalid_Operation_Flag := True;
   EvalLog := T;
End; {EvalExp}

{----------------------------------------------------------------------------}
Function Evaluate(Func: FuncPtr; OfType: TypeOfFunc; X: Float) : Float;
{ a semi-general evaluation function }

Var T, T2: Float;
Begin
    T:=0;
    T2:=0;

    Case OfType of
         Pol: T:=EvalPolynom(Func^.Poly,X);

         Rat:
            Begin
               T:=EvalPolynom(Func^.Ratio.Num,X);
               T2:=EvalPolynom(Func^.Ratio.Denom,X);

               If T2=0 then Invalid_Operation_Flag := True {Division by zero !}
                       else T := T/T2;
            End;
         Exp: T := EvalExp(Func,X);
         Log: T := EvalLog(Func,X);
       End;
    Evaluate:=T;
End;

{----------------------------------------------------------------------------}
Procedure SetRoot(Var Poly: PolyType; Root: Complex);
{ Allocate memory and store Root into root list of Poly }

Var Index: RootPtr;
    i: Integer;
Begin
   If Poly.RootCount=0 then
      New(Poly.Roots);

   Index:=Poly.Roots;
   For i:=2 to Poly.RootCount do
       Index:=Index^.Link;

   If Poly.RootCount>0 then
      Begin
         New(Index^.Link);
         Index:=Index^.Link;
      End;

   Index^.Root:=Root;
   Index^.Link:=NIL;
   Inc(Poly.RootCount);
End; {SetRoot}

{----------------------------------------------------------------------------}
Procedure GetRoot(Poly: PolyType; N: Byte; Var Result: Complex);
{ Returns the N'th zero of a polynomial (assuming it's in stack) }
Var Index: RootPtr;
    i: Integer;
Begin
   Index:=Poly.Roots;
   For i:=1 to N do
       Index:=Index^.Link;

   Result:=Index^.Root;
End; {GetRoot}

{----------------------------------------------------------------------------}
Function RootExist(Poly: PolyType; Z: Complex): Boolean;
{ Check if Z is in Roots Stack }
Var Index: RootPtr;
    Found: Boolean;
    I: Byte;
Begin
    Index:=Poly.Roots;
    I:=0;
    Found := False;
    While (Not Found) and (I<Poly.RootCount) do
      Begin

        With Index^ do
             Found:= (Root.Re=Z.Re) and (Root.Im=Z.Im);
        Index:= Index^.Link;
        Inc(I);
      End;

    RootExist:=Found;
End; {RootExist}

{----------------------------------------------------------------------------}
Function GetCoeff(Poly: PolyType; Exp: DegreeType): Float;
{ Returns the coeff. of the X^Exp in Poly }
Var Index: PolyPtr;
    i: Integer;
Begin
    Index:=Poly.Polynom;
    i:=0;
    If Exp>0 then
       For i:=1 to Exp do
           Index:=Index^.Link;

    GetCoeff:=Index^.Coeff;
End; {GetCoeff}

{----------------------------------------------------------------------------}
Procedure Polynom_Copy(Source: PolyType; Var Dest: PolyType);
{ Copies polynom. Source to Dest (Excluding roots) }
Var i: Byte;
    Index, indexSrc : PolyPtr;

Begin
   New(Dest.Polynom);
   Index:=Dest.Polynom;
   IndexSrc:=Source.Polynom;
   If Source.Exp > 0 then
      For i:=0 to Source.Exp-1 do
          Begin
             Index^.Coeff := IndexSrc^.Coeff;
             New(Index^.Link);
             Index:=Index^.Link;
             IndexSrc:=IndexSrc^.Link;
          End;
   Index^.Coeff := IndexSrc^.Coeff;
   Index^.Link := Nil;
   Dest.Exp := Source.Exp;
End; {Polynom_Copy}

{----------------------------------------------------------------------------}
Procedure Polynom_Dispose(Var Poly: PolyType);
{ Dispose memory allocated to polynomial Poly.Polynomial
  (excluding memory allocated for the roots) }
Var i: Byte;
    Index,Next : PolyPtr;
Begin
    Index:=Poly.Polynom;
    For i:=0 to Poly.Exp do
        Begin
           Next:=Index^.Link;
           Dispose(Index);
           Index:=Next;
        End;
    Poly.Exp:=0;
End;

{----------------------------------------------------------------------------}
Procedure Get(Var Func: FuncPtr);
Var StrFunc: String;

Function Parse_Polynom(Var Poly: PolyType; StrFunc: String): Boolean;
Var i, Items: Integer;
    V, VIdx, Dis: PolyHelpPtr;
    Sign: SignType;
    DegStr, CoeffStr: String;
    Is_OK, Variable, DegFlag: Boolean;
    Max: DegreeType;

Procedure Put(Var PIdx: PolyHelpPtr; DegS, CoeffS: String; Sign: SignType);

Var ERR1,ERR2: Integer;
    Coeff: Float;
    Degree: DegreeType;
    OKAY: Boolean;
Begin
    Val(DegS,Degree,ERR1);
    Val(CoeffS,Coeff,ERR2);
    OKAY := (ERR1=0) and (ERR2=0) and (Degree <= MaxDegree);
    If (Okay) and (Is_OK) then
       Begin
          If VIdx=Nil then
             Begin
                New(V);
                VIdx:=V;
             End
          else
             Begin
                New(VIdx^.Link);
                VIdx:=Vidx^.Link;
             End;

          VIdx^.Coeff := Sign*Coeff;
          VIdx^.Deg   := Degree;
          VIdx^.Link  := Nil;
          Inc(Items);

          If Degree > Max then Max:=Degree;
       End;

    Is_OK:=OKAY;
End; {Put}

{----------------------------------------------------------------------------}
Procedure Process;
Begin
   If CoeffStr='' then CoeffStr:='1.0';
   If DegStr  ='' then
       Begin
          If Variable then DegStr:='1'
                      else DegStr:='0';
       End;
   Put(V,DegStr,CoeffStr,Sign);
   CoeffStr :='';
   DegStr   :='';
   Variable:=False;
   DegFlag:=False;
End; {Process}

{----------------------------------------------------------------------------}
Procedure Sort_List;
Var j, Top: Integer;
    Idx   : PolyHelpPtr;
    TFloat: Float;
    TByte : Byte;
    Swap: Boolean;
Begin
   Top := Items;
   Repeat
       Swap:=False;
       Idx:=V;
       For j:=1 to Top-1 do
           If Idx^.Deg > Idx^.Link^.Deg then
              Begin
                 TFloat           := Idx^.Coeff;
                 Idx^.Coeff       := Idx^.Link^.Coeff;
                 Idx^.Link^.Coeff := TFloat;

                 TByte          := Idx^.Deg;
                 Idx^.Deg       := Idx^.Link^.Deg;
                 Idx^.Link^.Deg := TByte;

                 Idx := Idx^.Link;
                 Swap:=True;
              End;
       Dec(Top);
   Until Not Swap;
End; {Sort_List}

{----------------------------------------------------------------------------}
Var
  ProcessIt: Boolean;
  Index: PolyPtr;
  Len: Byte;
Begin
   Is_OK:=True;
   Sign  :=1;  {+}
   CoeffStr :='';
   DegStr   :='';
   Variable:=False;
   DegFlag:=False;
   ProcessIt:=False;
   Max:=0;
   Len := Length(StrFunc);
   Items := 0;
   VIdx:=Nil; {Mark start of linked-list}

   i:=1;
   While (Is_Ok) and (i<=Len) do
   Begin
       If Not (StrFunc[i] in ['-','+']) then ProcessIt:=True;
       Case StrFunc[i] of
        '+','-':
              Begin
                 If ProcessIt then Process;
                 If StrFunc[i]='+' then Sign:=1
                                   else Sign:=-1;
              End;
        ' ': { Do nothing ... Just skip it } ;
        'x','X':
               Variable:=True;

            '^':
               If Variable then
                             DegFlag := True
                           else
                             Is_OK := False;
  '0'..'9','.':
              If DegFlag then
                           Begin
                              If StrFunc[i]='.' then Is_OK:=False
                                                else DegStr:=DegStr+StrFunc[i]
                           End
                         else
                           CoeffStr:=CoeffStr+StrFunc[i];
         else
           Is_OK := False;
       End;

       Inc(i);
   End;

   If (Len>0) and (Is_Ok) then Process
                          else Is_OK := False;

   If Is_OK then
      Begin
         Poly.Exp:=Max;
         If (Items>1) and (Poly.Exp>0) then Sort_List;
         New(Poly.Polynom);
         Index:=Poly.Polynom;
         VIdx :=V;
         If Poly.Exp>0 then
            For i:=0 to Max-1 do
                If i=VIdx^.Deg then
                   Begin
                      Index^.Coeff := VIdx^.Coeff;
                      New(Index^.Link);
                      Index:=Index^.Link;
                      Dis := VIdx;        {Store previous item }
                      VIdx := VIdx^.Link; {Point to next }
                      Dispose(Dis);       {Dispose of previous }
                   End
                else
                   Begin
                      Index^.Coeff := 0;
                      New(Index^.Link);
                      Index:=Index^.Link;
                   End;

         Index^.Coeff:= VIdx^.Coeff;
         Index^.Link := Nil; {End List}
         Dispose(VIdx);
      End
        else
          Begin
             GotoXY(8+i,WhereY);
             Write('^---------- Error');
          End;

   Parse_Polynom:=Is_OK;
End; {Parse_Polynom}

{----------------------------------------------------------------------------}
Function StripAll(Str: String): String;
{ Delete all space chars in a string }

Var Str2: String;
    i   : Integer;
Begin
   Str2:='';
   For i:=1 to Length(Str) do
       If Str[i]<>' ' then Str2:=Str2+Str[i];

   StripAll := Str2;
End; {StripAll}

{----------------------------------------------------------------------------}
Procedure Pol_Message;
Begin
    If Skip_Message then Exit;
    Writeln('    Polynomial function :');
    Writeln('    P(x) = [An]*x^(n) + [An-1]*x^(n-1) ++ [A1]*x + [A0]');
    Writeln;
    Writeln('    Of which A0..An should be replaced with the appropriate coefficients,');
    Writeln('    and the exponent n represents a natural number');
    Writeln('    (No significance for exponents order).');
    Writeln;
    Writeln('    The multiply token ''*'' needn''t be written,');
    Writeln('    and there''s no obligation to write coefficients which equal ''1''.');
    Writeln('    Blanks have no impact on the input.');
    Writeln;
    Writeln('    For example: G(X) = x^2 - 1');
    Writeln('             or  P(X) = x^3-6x^2+11x-6');
End; {Pol_Message}

Procedure GetIt;
Begin
   GotoXY(10,WhereY-1);
   While Not Keypressed do;
   ClrEol;
   GotoXY(1,WhereY+1);
   ClrEol;
   GotoXY(10,WhereY-1);
   Readln(StrFunc);
   If StrFunc='!' then
      Begin
         Quit;
         DisposeOf(Funcs);
         Halt;
      End;
End; {GetIt}

{----------------------------------------------------------------------------}
Function Parse_Exp(Func: FuncPtr; StrFunc: String): Boolean;

Var OK: Boolean;
    T: Byte;
    BaseStr, FreeStr: String;
    Err1, Err2: Integer;

Begin
   StrFunc := StripAll(StrFunc);
   T := Pos('^',StrFunc);
   OK := (T>0) and (StrFunc[T+1] in ['X','x']);
   If OK then
      Begin
         BaseStr := Copy(StrFunc,1,T-1);
         FreeStr := Copy(StrFunc,T+2,Length(StrFunc));
         Val(BaseStr,Func^.BaseE,Err1);
         If FreeStr = '' then
            Begin
               Func^.FreeE:=0;
               Err2:=0;
            End
         else
            Val(FreeStr,Func^.FreeE,Err2);

         Err1 := Ord( (Err1=0) and (Func^.BaseE>0) and (Func^.BaseE<>1));
         Err1 := Ord( Not Boolean(Err1));
         OK := (Err1=0) and (Err2=0);
      End;

   If Not OK then
      Begin
         If (T=0) or Not (StrFunc[T+1] in ['X','x']) then
            Write('  Error - Can''t find ^x token!')
         else
            If Err1 <> 0 then Write('  Error - invalid base!')
                         else Write('  Error - invalid free coeff.');
      End;

   Parse_Exp := OK;
End; {Parse_Exp}

Function Parse_Log(Func: FuncPtr; StrFunc: String): Boolean;

Var OK: Boolean;
    T: Byte;
    BaseStr, FreeStr: String;
    Err1, Err2: Integer;

Begin
   StrFunc := StripAll(UpStr(StrFunc));
   T := Pos(')', StrFunc);

   OK := ( Length(StrFunc)>=7 ) and (copy(StrFunc,1,5)='LOGX(' ) and (T>0);

   If OK then
      Begin
         BaseStr := Copy(StrFunc,6,T-Pos('(',StrFunc)-1);
         FreeStr := Copy(StrFunc,T+1,Length(StrFunc));
         Val(BaseStr,Func^.BaseL,Err1);
         If FreeStr = '' then
            Begin
               Func^.FreeL:=0;
               Err2:=0;
            End
         else
            Val(FreeStr,Func^.FreeL,Err2);

         Err1 := Ord( (Err1=0) and (Func^.BaseL>0) and (Func^.BaseL<>1));
         Err1 := Ord( Not Boolean(Err1));
         OK := (Err1=0) and (Err2=0);
      End;

   If Not OK then
      Begin
         If (T=0) or (Length(StrFunc)<7) then
            Write('  Error - Can''t find ''log x(a)'' token !')
         else
            If Err1 <> 0 then Write('  Error - invalid base !')
                         else Write('  Error - invalid free coeff. ! ');
      End;

   Parse_Log := OK;
End; {Parse_Log}
{----------------------------------------------------------------------------}
Begin
   ClrScr;
   Writeln;
   Writeln('  Please enter the required function -');
   Writeln;
   Case Func^.TypeOf of
        Pol:
           Begin
              Pol_Message;
              Writeln;
              Writeln('  F(x) = ');
              Repeat GetIt;
              Until Parse_Polynom(Func^.Poly,StrFunc)=True;
          End;

        Rat:
           Begin
              Writeln('      Rational function :');
              Writeln('              P(x)');
              Writeln('       R(x) = ----');
              Writeln('              G(x)');
              Writeln;
              Writeln('    Of which P(X) and G(X) are polynomials.');
              Pol_Message;
              Writeln;
              Writeln('  P(x) = ');
              Repeat GetIt;
              Until Parse_Polynom(Func^.Ratio.Num,StrFunc)=True;

              Writeln('  G(x) = ');
              Repeat GetIt;
              Until Parse_Polynom(Func^.Ratio.Denom,StrFunc)=True;
           End;
        Exp:
            Begin
               Writeln('        Exponential function : E(x) = A^x - B');
               Writeln;
               Writeln('        Of which the base A is positive and different from 1( A<0 and A<>1 )');
               Writeln;
               Writeln('  E(x) = ');
               Repeat GetIt;
               Until Parse_Exp(Func,StrFunc)=True;
            End;
        Log:
            Begin
               Writeln('       Logarithmic function : L(x) = log x(A)+B');
               Writeln;
               Writeln('       Of which the base A is positive and different from 1 ( A<0 and A<>1 )');
               Writeln;
               Writeln('  L(x) = ');
               Repeat GetIt;
               Until Parse_Log(Func,StrFunc)=True;
            End;
   End;
   Skip_Message:=True;
End; {Get}

{----------------------------------------------------------------------------}
Function PolyStr(Poly: PolyType): String;
Var T: String;

Procedure Add(SubStr: String);
Begin
   T := T + SubStr;
End;

Var Idx: PolyPtr;
    V: HelpType;
    i: Integer;

Procedure Process;
Var SCoeff: String;
Begin
   If V[i] <> 0 then
      Begin
         If (V[i] > 0) and (i<>Poly.Exp) then Add('+');
         SCoeff:=RealF(V[i]);
         Case i of
              0: Add(SCoeff);
              1: If (SCoeff<>'1') and (SCoeff<>'-1') then Add(SCoeff+'x')
                    else
                       Begin
                          If SCoeff='-1' then Add('-');
                          Add('x');
                       End;
           else
              If SCoeff='1' then
                          Add('x^'+Int2Str(i) )
                        else If SCoeff='-1' then
                                          Add('-x^'+Int2Str(i) )
                                        else
                                          Add(SCoeff+'x^'+Int2Str(i) )
         End;
      End;
End; {Show}

Begin
   Idx := Poly.Polynom;
   For i:=0 to Poly.Exp do
       Begin
           V[i] := Idx^.Coeff;
           Idx := Idx^.Link;
       End;

   T := '';
   For i:=Poly.Exp downto 0 do
       Process;

   If T='' then T:='0';
   PolyStr := T;
End; {PolyStr}

{----------------------------------------------------------------------------}
Procedure CheckPoly(Poly: PolyType);
Var RootIdx : RootPtr;
    PolyIdx : PolyPtr;
    i: Integer;
Begin
   RootIdx := Poly.Roots;
   i:=0;
   While (RootIdx<>Nil) and (i<MaxDegree) do
         Begin
            Inc(i);
            RootIdx := RootIdx^.Link;
         End;
   Write('Roots: ');
   If i=Poly.RootCount then Writeln('Ok')
                       else Writeln('!!');

   PolyIdx := Poly.Polynom;
   i:=-1;
   While (PolyIdx<>Nil) and (i<MaxDegree) do
         Begin
            Inc(i);
            PolyIdx := PolyIdx^.Link;
         End;
   Write('Polynom:');
   If i=Poly.Exp then Writeln('Ok')
                 else Writeln('!!');

   Writeln;
End; {CheckPoly}

{----------------------------------------------------------------------------}
Procedure DisposePoly(Var Poly: PolyType);
{ Dispose memory allocated to coeff. & roots of polynomial Poly.Polynom }
Var RootIdx, NextRoot: RootPtr;
    PolyIdx, NextPoly : PolyPtr;
    Var i: Integer;
Begin
   RootIdx := Poly.Roots;
   While RootIdx <> Nil do
         Begin
            NextRoot := RootIdx^.Link;
            Dispose(RootIdx);
            RootIdx := NextRoot;
         End;
   Poly.Roots:=Nil;

   PolyIdx := Poly.Polynom;
   While PolyIdx <> Nil do
         Begin
            NextPoly := PolyIdx^.Link;
            Dispose(PolyIdx);
            PolyIdx := NextPoly;
         End;
   Poly.Polynom:=Nil;
End; {DisposePoly}

{----------------------------------------------------------------------------}
Procedure Find_Poly_Signs(Poly: PolyType; Var BInfo: LongInt; Var V: HelpType;
          Var Tail: Integer);

Var Temp: Float;
    Found: Boolean;

Procedure Search_For_Duplicates;
Var i: Integer;
Begin
   i:=1;
   Found := False;

   While (i<Tail) and (Not Found) do
         Begin
            Found := (V[Tail]=V[i]);
            Inc(i);
         End;
End; {Search_For_Duplicates}

Procedure Bubble_Sort;
Var i,j,k: Integer;
Begin
   k := Tail-1;
   For j:=1 to Tail-2 do
       Begin
         For i:=1 to k-1 do
             If V[i]>V[i+1] then {Swap Them}
             Begin
                Temp   := V[i+1];
                V[i+1] := V[i];
                V[i]   := Temp;
             End;
         Dec(K);
       End;
End; {Bubble_Sort_Step}

Var i: Integer;
    Index: RootPtr;
Begin
  { Bit Number / (True value) states that the function                      }
  { -------------------------------------------------                       }
  { 0          - Is defined for every (Real) X                              }
  { 1          - Is defined when X=0                                        }
  { 2          - Has got an horizontal asymptote                            }
  { 3          - Has got a curved asymptote                                 }
  { 4..7 => n  - Number of distinct derivative zero's (and undefined values)}
  { 8..8+n     - Gives a postive values (+) on the matching interval.       }

   ClearVec(V);
   Index := Poly.Roots;
   Tail := 1;

   If Poly.RootCount=0 then { No zeros so check at x=0 }
      Begin
        If EvalPolynom(Poly,0) > 0 then
           BInfo := SetBit(BInfo,8);
      End
   else
     Begin
        For i:=1 to Poly.RootCount do
            Begin
               If Abs(Index^.Root.Im)>EPSILON*1e10 then
                  Begin
                     Index:=Index^.Link;
                     Continue;
                  End;

               V[Tail] := Index^.Root.Re;
               Search_For_Duplicates;
               If Not Found then Inc(Tail);
               Index := Index^.Link;
            End;

        Bubble_Sort;

        For i:=1 to Tail-1 do
            Begin
               If i=1 then Temp:=V[1]-10
               else Temp:=( (v[i]+v[i-1]) / 2 );

               If EvalPolynom(Poly,Temp) > 0 then
                  BInfo := SetBit(BInfo,i+7);
            End;

        If EvalPolynom(Poly,v[1]+10) > 0 then
           BInfo := SetBit(BInfo,(Tail-1)+8);
     End;
End; {Find_Poly_Signs}

{----------------------------------------------------------------------------}
Procedure Find_Ratio_Signs(Ratio: RatioType; Var BInfo: LongInt; Var V: HelpType;
          Var Tail: Integer);
Var V1,V2: HelpType;
    T1,T2: Integer;
    Inf1,Inf2: LongInt;
Begin
   Find_Poly_Signs(Ratio.Num,Inf1,V1,T1);
   Find_Poly_Signs(Ratio.Denom,Inf2,V2,T2);

   { }
End;
{----------------------------------------------------------------------------}
Function C_Write(Z: Complex): String;
Begin
   { }

   If Z.Im=0 then C_Write := RealF(Z.Re)
             else C_Write := RealF(Z.Re) + ' + ' + RealF(Z.Im) + 'i';
End;

{$IFDEF VERBOSE}
Function C_Write_(Z: Complex): String;
Begin
   { }

   If Z.Im=0 then C_Write_ := RealF_(Z.Re)
             else C_Write_ := RealF_(Z.Re) + ' + ' + RealF_(Z.Im) + 'i';
End;
{$ENDIF}

Procedure Analyze(Func: FuncPtr);

Procedure AddItem(Var MInfo: MInfoPtr; X: Float);
Var Index: MinfoPtr;
    i: Byte;
Begin
   If MInfo=Nil then
      Begin
         New(MInfo);
         Index:=MInfo;
      End
    else
      Begin
         Index := MInfo;
         While Index^.Link<>Nil do
               Index:=Index^.Link;

         New(Index^.Link);
         Index:=Index^.Link;
      End;

   Index^.Math := X;
   Index^.Link := Nil;
End;

Function C_Muller(Var Poly: PolyType): Boolean;
{ Seek for an approx. of a zero of polynomial Poly }
{ Note Copyright (C) in README.TXT }

Procedure DividePoly(Var Poly: PolyType; Root: Float);
{ Perform a synthetic division of Poly in X minus found root }
{ Poly <-- Poly / (x-R) While R is the root }

Var i: Byte;
    Quotient, Vec: HelpType;
    Index: PolyPtr;
Begin
    For i:=0 to MaxDegree do
        Begin
            Vec[i]:=0;
            Quotient[i]:=0;
        End;

    Index:=Poly.Polynom;

    For i:=0 to Poly.Exp do
        Begin
            Vec[i]:=Index^.Coeff;
            Index:=Index^.Link
        End;

    { Vec[0]=Modulo }
    For i:=Poly.Exp downto 1 do
        Begin
          Quotient[i-1]:=Vec[i];
          Vec[i]:=0;
          Vec[i-1]:=Vec[i-1]+(Root*Quotient[i-1]);
        End;

    Index:=Poly.Polynom;
    For i:=0 to Poly.Exp-2 do
        Begin
            Index^.Coeff:=Quotient[i];
            Index:=Index^.Link;
        End;

    Index^.Coeff:=Quotient[i+1];
    Dispose(Index^.Link);
    Index^.Link:=Nil;

    Dec(Poly.Exp);
End; {DividePoly}

Procedure DividePoly2(Var Poly: PolyType; Coeff1,Coeff2: Float);
{
 Perform a synthetic division of Poly in (X-z)(x+z) (Z are the complex roots)
 Poly <-- Poly / (x^2-2ax+a^2+b^2) ; While Z=a+bi.
}

Var i: Byte;
    Quotient, Vec: HelpType;
    Index,Index2: PolyPtr;
Begin
    For i:=0 to MaxDegree do
        Begin
            Vec[i]:=0;
            Quotient[i]:=0;
        End;

    Index:=Poly.Polynom;

    For i:=0 to Poly.Exp do
        Begin
            Vec[i]:=Index^.Coeff;
            Index:=Index^.Link
        End;

    { Vec[0]=Modulo }
    For i:=Poly.Exp downto 2 do
        Begin
          Quotient[i-2]:=Vec[i];
          Vec[i]:=0;
          Vec[i-2]:=Vec[i-2]-(Coeff2*Quotient[i-2]);
          Vec[i-1]:=Vec[i-1]-(Coeff1*Quotient[i-2]);
        End;

    Index:=Poly.Polynom;
    For i:=0 to Poly.Exp-3 do
        Begin
            Index^.Coeff:=Quotient[i];
            Index:=Index^.Link;
        End;

    Index^.Coeff:=Quotient[i+1];
    Index2 := Index^.Link;
    Dispose(Index2^.Link);
    Dispose(Index^.Link);
    Index^.Link:=Nil;

    Dec(Poly.Exp,2);
End; {DividePoly}

Const _E1 = 4.5e1 {45};
                    { Requested difference between last 2 approx. (x),
                      relative to EPSILON :
                      abs(x[i+1]-x[i]) / abs(x[i] <= E1*Epsilon

                      AKA 'relative error tolerance on x[i]'
                    }


      _E2 = 5e1 {50};
                    { Requested Zeros precision (y),
                      relative to EPSILON : abs(y) <= E2*Epsilon

                      AKA 'error tolerance on f( x[i] )'
                    }


      _E3 = _E2*_E2*_E2*_E2*_E2*_E2 {E2^6};

                    { Acceptance threshold of a zero which does not satisfy
                      one of the above conditions (y; needed when we reach
                      MaxLoops iterations),
                      relative to EPSILON : abs(y) <= E3*Epsilon
                    }

      MaxLoops = 60;{ Max Iterations }

      DelayMS = 350; {Delay under Verbose Mode (In Borland's Milliseconds) }

Var x,fx,h,L: Array[1..4] of Complex;
    i, j: Integer;
    g,te1,te2,UnderRoot: Complex;
    SolutionOK, NoProgress: Boolean;

{ The numbers in parenthesis, (x), represent the equivalent part in
  the original algorithm }

Procedure Init(A1,A2,A3: Float);
{A1,A2,A3 are 3 approximations to the zero}
Var T: Complex;
Begin
   { 0.5,-0.5,0 - Original}
   CMake(A1,0,x[3]);  C_EvalPolynom(Poly,x[3],fx[3]);
   CMake(A2,0,x[2]);  C_EvalPolynom(Poly,x[2],fx[2]);
   CMake(A3,0,x[1]);  C_EvalPolynom(Poly,x[1],fx[1]);
{2}CSub(x[3],x[2],h[3]);

   CSub(x[2],x[1],T);
   CDiv(h[3],T,L[3]);
End; {Init}

Var t1,t2,t3,t4: complex;
    E1,E2,E3,a2,pyth: float;
Begin
   i:=0;                 { Init. Iteration counter }

   E1 := _E1*EPSILON;
   E2 := _E2*EPSILON;
   E3 := _E3*EPSILON;

   Init(1,-1,0);         { Set first three points/approximations }

   {$IFDEF VERBOSE}
   Writeln(i,':');
   Delay(DelayMS);
   Writeln('x=',C_Write_(x[3]),'; y=',C_Write_(fx[3]));
   Writeln;
   {$ENDIF}

   SolutionOK := CMag(fx[3]) <= e2;
   NoProgress := False;

   While (I<=MaxLoops) and (Not NoProgress) and (Not SolutionOk) do
{ Perform while :                                                       }
{ 1. We haven't exceeded the number of max. iterations - MaxLoops AND   }
{ 2. The ratio of the difference between last approx. to the previous one
     to the last approx. isn't smaller than the requested ratio. E1 AND }
{ 3. Last approx.'s Y value Magnitude is bigger than E2                           }

   Begin
{3}   CMake(2,0,T1);
      CMult(T1 ,L[3],T1);
      CMake(1,0,T2);
      CAdd(T1,T2,T1);
      { T1 <-- 1+2*L[3]}

      CSub(fx[3],fx[2],T3);
      CMult(T1,T3,T1);
      { T1 <-- (1+2*L[3]) * (fx[3]-fx[2]) }

      CSqr(L[3],T2);
      CSub(fx[2],fx[1],T3);
      CMult(T2,T3,T2);
      { T2 <-- L[3]*L[3]- (fx[2]-fx[1])  }

      CSub(T1,T2,T1);
      G:=T1;

{ * The pascal code in remark is the old code for REAL numbers & roots }

{ *   g := (1+2*L[3])*(fx[3]-fx[2]) - (L[3]*L[3]*(fx[2]-fx[1]));}

{4}   CMult(T3,L[3],T1);
      { T1 <-- L[3]*(fx[2]-fx[1]) }
      CSub(fx[3],fx[2],T2);
      { T2 <-- fx[3]-fx[2] }
      CSub(T2,T1,T2);
      { T2 <-- T2-T1 }
      CMult(T2,L[3],T2);
      CAdd(COne,L[3],T1);
      CMult(T1,T2,T1);
      CMult(T1,fx[3],T1);
      CMake(4,0,T2);
      CMult(T2,T1,T2);
      CSqr(g,T1);
      CSub(T1,T2,T1);
      UnderRoot := T1;

{ *    UnderRoot:=g*g - 4*fx[3]*(1+L[3])*L[3]*(fx[3]-fx[2]-L[3]*(fx[2]-fx[1]));}
{ *    If UnderRoot.Re < 0 then UnderRoot:=CZero; }
      CSqrt(UnderRoot,T1);
      CAdd(g,T1,te1);
      CSub(g,T1,te2);
{ *   te1:=g+Sqrt(UnderRoot);
  *   te2:=g-Sqrt(UnderRoot);    }
      If CMag(te2)>CMag(te1) then te1:=te2;
      If (te1.Re=0) and (te1.Im=0) then te1:=COne; {Seems necessary but what
      is the meaning of this ? }
      CAdd(COne,L[3],T3);
      CMult(T3,fx[3],T3);
      CMake(-2,0,T2);
      CMult(T2,T3,T1);
      CDiv(T1,te1,L[4]);
{ *   L[4]:=(-2*fx[3]*(1+L[3])) / t1;}

{5}   CMult(h[3],L[4],T1);
      CAdd(T1,x[3],x[4]);
{ *   x[4]:=x[3]+h[3]*L[4];}

      h[4] := T1;
{ *   h[4]:=h[3]*L[4];}

{6}   C_EvalPolynom(Poly,x[4],fx[4]);
      For j:=1 to 3 do {Rotate down : 4>>3>>2>>1>>Out }
          Begin
             x [j] := x [j+1];
             fx[j] := fx[j+1];
             h [j] := h [j+1];
             L [j] := L [j+1];
          End;

      Inc(i);

      {$IFDEF VERBOSE}
      Writeln(i,':');
      Delay(DelayMS);
      Writeln('x=',C_Write_(x[3]),'; y=',C_Write_(fx[3]));
      { Note the thorem that states that if Z is a root of a polynomial
        with real coefficients then Conjugate(Z) is also a root }

      a2 := 2*x[3].Re;
      Pyth := Sqr(X[3].Re)+Sqr(X[3].Im);
      Writeln('Complex pair = x^2 - ',RealF_(a2),'x + ',RealF_(Pyth));
      {$ENDIF}

      CSub(x[3],x[2],T1);
      NoProgress := ( CMag(T1) / CMag(x[3]) ) <= e1 ;
      If (NoProgress) and (i<10) and (CMag(fx[3]) > e3) then
         { Too soon to quit, let's try with different starting values }
         Begin
            Init(3,1,2);
            NoProgress:=False;
            {$IFDEF VERBOSE}
            Writeln('Retrying...');
            {$ENDIF}
         End;

      If abs(x[3].Im) >= E3 then
         Begin
            CConj(x[3],T1);
            C_EvalPolynom(Poly,T1,T2);

            SolutionOK := ( CMag(fx[3])<=e2 ) and (CMag(T2)<=e2);
          End
      else
         SolutionOK := CMag(fx[3]) <= e2;

      {$IFDEF VERBOSE}
      Writeln;
      {$ENDIF}
   End;

   {$IFDEF VERBOSE}
   Writeln('i=',i);
   Writeln('I<=MaxLoops: ',I<=MaxLoops);
   Writeln('NoProgress: ',NoProgress);
   Writeln('SolutionOK: ',SolutionOK);
   Delay(6*DelayMS);
   Writeln;
   {$ENDIF}

   If CMag(fx[3]) <= e3 then
      Begin
         If abs(x[3].Im) >= e3 then
         Begin

            {$IFDEF VERBOSE}
            If (CMag(fx[3])<=e2) and (CMag(T2)>e2+e2) then
            begin
               writeln('impossible : no f(x) symmetry (call 911) ');
               halt;
            end;
            {$ENDIF}

            CConj(x[3],T1);        { T1 Should contain it anyway }
            SetRoot(Poly,x[3]);
            SetRoot(Poly,T1);
            a2 := 2*x[3].Re;
            Pyth := Sqr(X[3].Re)+Sqr(X[3].Im);
            DividePoly2(Poly, -a2 , Pyth);
            C_Muller:=True;
         End
          else
          Begin
             SetRoot(Poly,x[3]);
             DividePoly(Poly,x[3].Re);
             C_Muller:=True;
          End
      End
   else
      C_Muller:=False;
End; {Muller}

{----------------------------------------------------------------------------}
Procedure FindZeros(Var Poly: PolyType);

Procedure SolveQuad;
Var Root1,a,b,c,Delta: Float;

    SqrtDelta,Temp,Temp2,Temp3: Complex;

Begin
    a:=GetCoeff(Poly,2);
    b:=GetCoeff(Poly,1);
    c:=GetCoeff(Poly,0);
    Delta:=b*b-4*a*c;
    If Delta>=0 then Poly.NoRealRoots:=False;
    If Delta=0 then
       { Two identical real roots }
       Begin
          Root1:=-B/(2*a);
          CMake(Root1,0,Temp);
          SetRoot(Poly,Temp);
          SetRoot(Poly,Temp);
       End
    else {If Delta > 0 then }
       { Two different real roots }
       Begin
           { Ugly tests have shown no difference in accuracy while using
             the 2nd formula for quadratic roots
             OR calculating Root2 independentaly of Root1.
           }

           CMake(Delta,0,Temp);
           CSqrt(Temp,SqrtDelta);
           {SqrtDelta <-- CSqrt(Delta);}

           CMake(b,0,Temp2);
           CSub(SqrtDelta,Temp2,Temp2);

           CMake(a,0,Temp);
           CMake(2,0,Temp3);
           CMult(Temp,Temp3,Temp3);

           CDiv(Temp2,Temp3,Temp3);
           { Root1 <-- (-b+SqrtDelta) / (2*a);}
           SetRoot(Poly, Temp3);

           CDiv(SqrtDelta,Temp,Temp);
           CSub(Temp3,Temp,Temp);
           SetRoot(Poly, Temp {Root1 - (SqrtDelta/a) });
       End;
End; {SolveQuad}

{----------------------------------------------------------------------------}
Procedure Div_By_X(Var Poly: PolyType);
Var Index,Next: PolyPtr;
    i: Byte;
Begin
   Index := Poly.Polynom;
   For i:=0 to Poly.Exp-1 do
       Begin
          Next:=Index^.Link;
          Index^.Coeff:=Next^.Coeff;
          If (i<>Poly.Exp-1) then Index:=Next;
       End;
   Dispose(Index^.Link);
   Index^.Link:=Nil;
   Dec(Poly.Exp);
End; {Div_By_X}

{----------------------------------------------------------------------------}
Var  PolyBack: PolyType;
     RootFound: Boolean;
     Temp:    Complex;
     OldExp : DegreeType;
Begin
   Polynom_Copy(Poly,PolyBack);
   While (Poly.Polynom^.Coeff=0 {Free Coeff=0}) and (Poly.Exp>2) do
         Begin
            Div_By_X(Poly);
            SetRoot(Poly,CZero);
         End;

   RootFound:=True;
   While (Poly.Exp > 2) and (RootFound) do
         Begin
            {$IFDEF VERBOSE}
            Writeln('Working on: ',PolyStr(Poly));
            {$ENDIF}
            OldExp := Poly.Exp;
            RootFound := C_Muller(Poly);
            If Poly.Exp=OldExp-1 then Poly.NoRealRoots:=False;
         End;

   {$IFDEF VERBOSE}
   Writeln('Working on: ',PolyStr(Poly));
   Writeln('Press <ENTER> to continue.');
   Readln;
   {$ENDIF}
   Case Poly.Exp of
        2: SolveQuad;
        1:
         Begin
            CMake(-GetCoeff(Poly,0) / GetCoeff(Poly,1) , 0, Temp);
            SetRoot(Poly,Temp);
            Poly.NoRealRoots := False;
{        SetRoot(Poly, -GetCoeff(Poly,0) / GetCoeff(Poly,1));}
         End;
        0: If GetCoeff(Poly,0)=0 then
           Begin
              SetRoot(Poly, CZero);
              Poly.NoRealRoots:=False;
           End;
   End;
   Polynom_Dispose(Poly);
   Polynom_Copy(PolyBack,Poly);
   Polynom_Dispose(PolyBack);
End; {FindZeros}

{----------------------------------------------------------------------------}
Procedure SubPoly(A,B: PolyType; Var Diff: PolyType);
{Diff := A-B}
Function Max(A,B: LongInt): LongInt;
Begin
   If A>B then Max:=A
          else Max:=B;
End;

Var Index: PolyPtr;
    I, Big: Byte;
    V: HelpType;
Begin
   For i:=0 to MaxDegree do
       V[i] := 0;

   Big:=0;
   Index := A.Polynom;
   For i:=0 to A.Exp do
       Begin
          V[i] := V[i] + Index^.Coeff;
          Index := Index^.Link;
       End;

   Index := B.Polynom;
   For i:=0 to B.Exp do
       Begin
          V[i] := V[i] - Index^.Coeff;
          Index := Index^.Link;
       End;

   i := Max(A.Exp,B.Exp);
   While (i>0) and (V[i]=0) do Dec(i);  { Find the exp. of the difference. polynom. }
   Big := i;

   New(Diff.Polynom);
   Index := Diff.Polynom;
   If Big>0 then
      For i:=0 to Big-1 do
          Begin
             Index^.Coeff := V[i];
             New(Index^.Link);
             Index := Index^.Link;
          End;
   Diff.Exp:=Big;
   Index^.Coeff := V[Big];
   Index^.Link := Nil;
End; {SubPoly}

{----------------------------------------------------------------------------}
Procedure MultPoly(A,B: PolyType; Var Product: PolyType);
Var i,j: Integer;
    Big: Byte;
    AIdx, BIdx: PolyPtr;
    V: HelpType;
Begin
   For i:=0 to MaxDegree do
       V[i] := 0;

   AIdx := A.Polynom;
   Big := 0;
   For i:=0 to A.Exp do
       Begin
          BIdx := B.Polynom;
          For j:=0 to B.Exp do
            Begin
               V[i+j]:=V[i+j]+(AIdx^.Coeff*BIdx^.Coeff);
               BIdx := BIdx^.Link;
               If V[i+j] <> 0 then Big := i+j;
            End;
          AIdx := AIdx^.Link;
       End;

   New(Product.Polynom);
   AIdx := Product.Polynom;
   If Big > 0 then
      Begin
         For i:=0 to Big-1 do
             Begin
                AIdx^.Coeff := V[i];
                New(AIdx^.Link);
                AIdx := AIdx^.Link;
             End;
        AIdx^.Coeff := v[i+1];
        Product.Exp := i+1;
      End
   else
      Begin
        AIdx^.Coeff := v[0];
        Product.Exp := 0;
      End;

   AIdx^.Link := Nil;
End; {MultPoly}

{----------------------------------------------------------------------------}
Procedure PolyDeriv(Var Poly: PolyType);
{ Using the rule : (X)' = n*X^(n-1)  }
Var Index,Next: PolyPtr;
    i: Byte;
Begin
   Index := Poly.Polynom;
   If Poly.Exp > 0 then
      Begin
         For i:=0 to Poly.Exp-1 do
             Begin
                Next:=Index^.Link;
                Index^.Coeff:=(i+1)*Next^.Coeff;
                If (i<>Poly.Exp-1) then Index:=Next;
             End;
        Dispose(Index^.Link);
        Index^.Link:=Nil;
        Dec(Poly.Exp);
      End
   else
      Index^.Coeff:=0;
End; {PolyDeriv}

{----------------------------------------------------------------------------}
Procedure RatioDeriv(Var RatioD: RatioType);
{                  [ f(X) ]'  f'(X)*g(X) - f(X)*g'(X) }
{ Using the rule : [ ---- ] = ----------------------- }
{                  [ g(X) ]           g(X)           }


Var NumDiff, DenomDiff, T1, T2: PolyType;

Begin
   Polynom_Copy(RatioD.Num  ,NumDiff);
   Polynom_Copy(RatioD.Denom,DenomDiff);

   PolyDeriv(NumDiff);                      {f'}
   PolyDeriv(DenomDiff);                    {g'}

   MultPoly(NumDiff   , RatioD.Denom, T1); {f'g}
   MultPoly(RatioD.Num , DenomDiff  , T2); {fg'}

   Polynom_Dispose(RatioD.Num);
   SubPoly(T1,T2,RatioD.Num);
   Polynom_Dispose(T1);
   Polynom_Dispose(T2);

   MultPoly(RatioD.Denom, RatioD.Denom, T1); {g}
   Polynom_Dispose(RatioD.Denom);
   Polynom_Copy(T1,RatioD.Denom);

   Polynom_Dispose(T1);
   Polynom_Dispose(NumDiff);
   Polynom_Dispose(DenomDiff);
End; {RatioDeriv}

{----------------------------------------------------------------------------}
Procedure Find_Asymptotes;
Var A : Integer;
    Asym, M, N: Float;
    AsymLim, AsymLim2, DenomX: PolyType;
Begin
   With Func^.Ratio do
        A := Num.Exp - Denom.Exp;

   With Func^.Ratio do Case A of
        1 : { Curved asym. }
          Begin
             With DenomX do
                  Begin
                     {Create the polynom. p(x) = x }
                     {Allocate DenomX}
                     New(Polynom);
                     New(Polynom^.Link);
                     Polynom^.Coeff:=0;
                     Polynom^.Link^.Coeff:=1;
                     Polynom^.Link^.Link:=Nil;
                     Exp:=1;
                  End;

             {AsymLim <-- Denom*X }
             MultPoly(Denom,DenomX,AsymLim);

             Func^.BInfo := SetBit(Func^.BInfo,3);
             M := GetCoeff(Num,Num.Exp) / GetCoeff(AsymLim,AsymLim.Exp);
             AddItem(Func^.MInfo,0); {2nd item is horizontal asym. }
             AddItem(Func^.MInfo,M); {3rd }

             Polynom_Dispose(AsymLim);

             DenomX.Polynom^.Link^.Coeff := M;   { DenomX = MX }
             MultPoly(Denom,DenomX,AsymLim);     { MX * Denom }
             SubPoly(Num,AsymLim,AsymLim2);      { Num - MX*Denom }

             if (AsymLim2.Exp=Denom.Exp)
                then N := GetCoeff(AsymLim2,AsymLim2.Exp) / GetCoeff(Denom,Denom.Exp)
             else
                if (AsymLim2.Exp<Denom.Exp)
                   then N := 0
                else {Can this happen?} RunError(201);

             AddItem(Func^.MInfo,N); {4th }

             Polynom_Dispose(AsymLim);
             Polynom_Dispose(AsymLim2);
             With DenomX do
                  Begin
                     {Dispose DenomX }
                     Dispose(Polynom^.Link);
                     Dispose(Polynom);
                  End;
          End;
  -255..0 :     { Horizon. Asym. : Na / Da }
          Begin
             Func^.BInfo := SetBit(Func^.BInfo,2);
             If A=0 then Asym := GetCoeff(Num,Num.Exp) / GetCoeff(Denom,Denom.Exp)
                    else Asym := 0;
             AddItem(Func^.MInfo,Asym);
          End;
   End;

End;

{----------------------------------------------------------------------------}
Var
  RootIdx: RootPtr;
  V: HelpType;
  Tail: Integer;

Begin {Analyze}
   ClrScr;

   With Func^ do
      Case TypeOf Of
        Pol:    { Polynomial }
           Begin
              BInfo := 3;   { Definition interval : All X ; X=0 Y-OK }

              FindZeros(Poly);               { Seek zeros      }

              Polynom_Copy(Poly, PolyDiff);  { Copy function   }

              PolyDeriv(PolyDiff);           { Calculate its derivative }
              FindZeros(PolyDiff);           { Seek derivative zeros }
              Find_Poly_Signs(PolyDiff, BInfo, V, Tail);{ Find derivative signs }

              AddItem(MInfo, Evaluate(Func,Pol,0));
              AddItem(MInfo, 0); { Horiz. Asym. }
           End;
        Rat:     { Rational }
           Begin
              FindZeros(Ratio.Num);            { Seek numerator zeros }
              FindZeros(Ratio.Denom);          { Seek denominator zeros }

              Polynom_Copy(Ratio.Num, RatioDiff.Num);
              Polynom_Copy(Ratio.Denom, RatioDiff.Denom);

              RatioDeriv(RatioDiff);           { Calculate its derivative }
              FindZeros(RatioDiff.Num);        { Seek derivative zeros  }
              FindZeros(RatioDiff.Denom);
              { Seek values for which the derivative isn't defined }

              {Find_Ratio_Signs(RatioDiff, BInfo, V, Tail);{ Find derivative signs }

              If Ratio.Denom.RootCount=0 then
                 BInfo := SetBit(BInfo,1);

              If Not RootExist(Ratio.Denom,CZero) then
                 Begin
                   BInfo := SetBit(BInfo,1);
                   AddItem(MInfo, Evaluate(Func,Rat,0));
                 End
              else
                 AddItem(MInfo,0);

              Find_Asymptotes;
           End;

        Exp:    { Exponential }
           Begin
              Func^.MaxX := 0;
              If (FreeE < 0) then
                 RootE := ln(-FreeE) / ln(BaseE)  { Root = a = }
              else
                 RootE := 0;

              BInfo := 7;
              AddItem(MInfo, EvalExp(Func,0));
              AddItem(MInfo,FreeE); { Horiz. Asym. }
           End;

        Log:    { Logarithmic }
           Begin
              Func^.MaxX := 0;
              RootL := 1 / pwrXY(BaseL,FreeL);

              BInfo := 1;
              AddItem(MInfo,0); { x=0 y-undefined }
              AddItem(MInfo,0); { Horiz. Asym.    }
           End;

      End;
End; {Analyze}

{----------------------------------------------------------------------------}
Function GetItem(MInfo: MinfoPtr; No: Byte): Float;
Var i: Byte;
    Index: MInfoPtr;
Begin
   Index := MInfo;
   For i:=1 to No-1 do
       Index:=Index^.Link;

   GetItem := Index^.Math;
End; {GetItem}

{----------------------------------------------------------------------------}
Procedure WriteAnalyze(Var Func: FuncPtr);

Var TempPoly: PolyType;
           V: HelpType;
      i,Tail: Integer;
   MaxX,MaxY: Float;

Procedure WritePolZeros(Pol: PolyType; Deriv: Boolean);
Var RootIdx: RootPtr;
          Y: Float;
    i: Byte;
    Complex_Root: Boolean;
Begin
   RootIdx:=Pol.Roots;
   {add .......    If Deriv then Find_Poly_Signs}
   For i:=1 to Pol.RootCount do
       Begin
          Complex_Root := Abs(RootIdx^.Root.Im)>100*EPSILON;
          If Options.Complex or not Complex_Root then
             Begin
               Write('  ',i,') x=',C_Write(RootIdx^.Root),' ');

               If not Complex_Root then {Complex_Roots aren't visible on our}
                  Begin                 {2-D Real Graph, thus aren't counted}
                                        {in the limits                      }
                     Y:=Evaluate(Func,Func^.TypeOf,RootIdx^.Root.Re);
                     If Deriv then Write('  y=',RealF(y));

                     If (Abs(RootIdx^.Root.Re)>MaxX) then MaxX:=Abs(RootIdx^.Root.Re);
                     If Abs(Y)>MaxY then MaxY:=Abs(Y);
                  End;

               RootIdx:=RootIdx^.Link;

               Writeln;
               If WhereY>22 then
                  Begin
                     Readln;
                     ClrScr;
                  End;
             End;
       End;
End; {WritePolZeros}

Var S1,S2: String;
    Temp : Complex;

Begin
   MaxX := 0;  { Init of the 'interesting interval' }
   MaxY := 0;
   ClrScr;
   Writeln;
   Case Func^.TypeOf of
        Pol:
           Write(' F(x) = ',PolyStr(Func^.Poly));

        Rat:
           Begin
              S1 := PolyStr(Func^.Ratio.Num);
              S2 := PolyStr(Func^.Ratio.Denom);
              i := BigInt(Length(S1),Length(S2));
              S1:=PutBetween(8,8+i,S1);
              S2:=PutBetween(8,8+i,S2);
              Writeln(S1);
              Write(' F(x) = ');
              For tail:=1 to i do Write('-');
              Writeln;
              Write(S2);
           End;
        Exp:
           Begin
              Write(' F(x) = ');
              Write(RealF(Func^.BaseE),'^x');
              If Func^.FreeE >0 then
                                  Write('+',RealF(Func^.FreeE))
                               else
                                  If Func^.FreeE <> 0 then
                                     Write(RealF(Func^.FreeE));
           End;

        Log:
           Begin
              Write(' F(x) = ');
              Write('log X(',RealF(Func^.BaseL),')');
              If Func^.FreeL >0 then
                                  Write('+',RealF(Func^.FreeL))
                               else
                                  If Func^.FreeL <> 0 then
                                     Write(RealF(Func^.FreeL));
           End;

   End;
   Writeln;
   Writeln;
   Writeln(' =================');
   Writeln;
   Write('  X=0 Y');
   If GetBit(Func^.BInfo,1)=1 then Writeln('=',RealF(GetItem(Func^.MInfo,1)))
                              else Writeln('-undefined');

   Writeln;
   Write('  Function''s zeros: ');

   Case Func^.TypeOf Of
        Pol:
           Begin
              TempPoly:=Func^.Poly;
              With TempPoly do
                   If RootCount<Exp then Write('(not all of them were identified)');
           End;
        Rat:
           Begin
              TempPoly:=Func^.Ratio.Num;
              With TempPoly do
                   If RootCount<Exp then Write('(not all of them were identified)');
           End;
        Exp: TempPoly.RootCount := Ord(Func^.FreeE < 0) ;
        Log: TempPoly.RootCount := 1;
   End;

   Writeln;

   If TempPoly.RootCount=0 then Writeln('   None !')
   else Case Func^.TypeOf of
             Exp: Writeln('  x=',RealF(Func^.RootE));
             Log: Writeln('  x=',RealF(Func^.RootL));
          else
             WritePolZeros(TempPoly,False)
         End;

   Writeln;
   Writeln;
   Case Func^.TypeOf of
        Exp:
           Begin
              S1 := RealF(Func^.BaseE);
              Write('  F''(x) = ','Ln(',S1,')*',S1,'^x');
           End;

        Log:
           Begin
              S1 := RealF(Func^.BaseL);
              Write('  F''(x) = ','1/[Ln(',S1,') * X',']');
           End;
        Pol:
           Write('  F''(x) = ',PolyStr(Func^.PolyDiff));

        Rat:
           Begin
              S1 := PolyStr(Func^.RatioDiff.Num);
              S2 := PolyStr(Func^.RatioDiff.Denom);
              i := BigInt(Length(S1),Length(S2));
              S1:=PutBetween(10,10+i,S1);
              S2:=PutBetween(10,10+i,S2);
              Writeln(S1);
              Write('  F''(x) = ');
              For tail:=1 to i do Write('-');
              Writeln;
              Writeln(S2);
           End;
   End;
   Writeln;
   If WhereY>15 then
      Begin
         Readln;
         ClrScr;
      End;

   Writeln;
   Writeln('  Derivative''s zeros :');

   { }
   Case Func^.TypeOf Of
        Pol: TempPoly:=Func^.PolyDiff;
        Rat: TempPoly:=Func^.RatioDiff.Num;
        Exp: TempPoly.RootCount := 0;
        Log: TempPoly.RootCount := 0;
   End;

   If TempPoly.RootCount=0 then
                             Writeln('   None !')
                           else
                             WritePolZeros(TempPoly,True);

   Writeln;
   Writeln('  Asymptotes:');
   Case Func^.TypeOf of
        Pol: Writeln('   None !');
        Exp:
           Begin
              Writeln('   Horizontal - ');
              Writeln('   y=',RealF(GetItem(Func^.MInfo,2)));
           End;

        Log:
           Begin
              Writeln('   Vertical -');
              Writeln('   x=0');
           End;

        Rat: With Func^.Ratio do
           Begin
              Writeln;
              Writeln('   Horizontal - ');
              If GetBit(Func^.BInfo,2)=0 then
                                           Writeln('   None.')
                                         else
                                           Writeln('   y=',RealF(GetItem(Func^.MInfo,2)));

              Writeln;
              Writeln('   Curved - ');
              If GetBit(Func^.BInfo,3)=0 then
                                           Writeln('   None.')
                                         else With TempPoly do
                                            Begin
                                               New(Polynom);
                                               New(Polynom^.Link);
                                               Polynom^.Coeff:=GetItem(Func^.MInfo,4);       {N}
                                               Polynom^.Link^.Coeff:=GetItem(Func^.MInfo,3); {M}
                                               Polynom^.Link^.Link:=Nil;
                                               Exp:=1;
                                               Writeln('   y=',PolyStr(TempPoly));
                                               Dispose(Polynom^.Link);
                                               Dispose(Polynom);
                                            End;

              Writeln;
              Writeln('   Vertical -');
              If (Denom.RootCount=0) or (Denom.NoRealRoots) then
                 Writeln('   None.')
                 Else
                   Begin
                      Find_Poly_Signs(Denom, Func^.BInfo, V, Tail);
                      For i:=1 to Tail-1 do
                          Begin
                             If Abs(V[i]) > MaxX then MaxX := Abs(V[i]);
                             Write('   ',i,') x=',RealF(V[i]));
                             CMake(v[i],0,Temp);
                             If RootExist(Num,Temp) then
                                Write(' (Removable discontinuity point) ');
                             Writeln;
                          End;
                   End;
           End;
   End;

   Func^.MaxX := MaxX;
   Func^.MaxY := MaxY;

   While Not Keypressed do;
   ReadKey;
End; {WriteAnalyze}

procedure SVGA_Mouse_Fix(Flags,CS,IP,AX,BX,CX,DX,SI,DI,DS,ES,BP : word);interrupt;
var hx,hy,hb,ev: Integer;
    hn,hm: integer;
    sx,sy,Smask,CMask, NewPixel: Integer;
begin
 ev:=ax;
 hx:=cx;
 hy:=dx;
 hb:=bx;

  if (hx<>mx) or (hy<>my) then
     begin
        sx := mx-HotX;
        sy := my-HotY;

        for hn:=0 to 15 do
            for hm:=0 to 15 do
                PutPixel(sx+hm,sy+hn,OldPixels[hm,hn]);

        sx := hx-HotX;
        sy := hy-HotY;

        For hn:=0 to 15 do
            For hm:=0 to 15 do
                OldPixels[hm,hn]:=Getpixel(sx+hm,sy+hn);

      for hn:=0 to 15 do
       for hm:=0 to 15 do
       Begin
          SMask := ColorCross[1,hm,hn];
          CMask := ColorCross[2,hm,hn];

          { The BW-Bitmap equivalent of following statement
            is more or less

            NewPixel := (OldPixels[hm,hn] and SMask) XOR CMask;
          }

          NewPixel := (White and CMask and SMask) XOR OldPixels[hm,hn];
          PutPixel(sx+hm,sy+hn, NewPixel);
       End;

      mx:=hx;
      my:=hy;
    end;

 inline ($8B/ $E5/ $5D/ $07/ $1F/ $5F/ $5E/ $5A/ $59/$5B/ $58/ $CB);
end;
{----------------------------------------------------------------------------}
Procedure Sketch(Funcs: FuncArr);

Var X0,Y0,M,N: Float;
    Newton: Boolean;

Function Max(A,B: Float): Float;
{ Returns the largest of the Real numbers }
Begin
    If A>B then
             Max:=A
           else
             Max:=B;
End; {Max}

Const
   Def_FuncColor = Cyan;
   Def_FGColor   = Yellow;
   Def_BGColor   = Blue;
   Def_AxisColor = White;

{  More defaults :

   Asymptotes <-- True;
   DrawLines  <-- True;
   GridInfo   <-- All;
   Axis       <-- Center;
}

Var
   T1,T2, XM,YM,X,Y,YMax,ScaleX,ScaleY: Float;
   MaxX, MaxY, MaxX_2, MaxY_2, MaxX_10, MaxY_10, XRes, YRes: Integer;
   Asymptotes, DrawLines, MatchScale, Quit: Boolean;
   FGColor, FuncColor, BGColor, AxisColor: Byte;
   MaxColors, BrackHeight, BrackWidth: Word;
   Axis: (Center,Frame,Hide);
   GridInfo: (All,Some,None);
   MouseInfo: (Buttom,Top,Nothing);
   FuncStr: 0..4;
   FLoc: Array[1..4] of PointType;
   Graphs: LongInt;

Function Res2Axis(N: Integer): Integer;
Begin
   With Options do
        If GraphFP=-1 then
        Begin
           N := N-2;
           If N<1 then N:=1;
        End
     else
        N := 1;

   Res2Axis := N;
End;

Procedure WriteXY(X,Y,N: Integer; St: Real);
Var T: String;
Begin
   Str(St:0:Res2Axis(N) ,T);
   OutTextXY(X,Y,T);
End;

Procedure InitMouse(MouseX,MouseY: Word);
Begin
   ResetMouse;
{   SetResolution(32,64);}
   SetMousePos(MouseX,MouseY);
   Set_GraphMouseCursor(BWCross, 8, 7);

   If Options.GDriver=SVGA256 then
      Begin
         { MaxX should be equal to GetMaxX+1
           MaxY should be equal to GetMaxY+1
         }

         { Check why doesn't Max-1 & MaxY-1 work. }
         SetRangeX(0,MaxX);
         SetRangeY(0,MaxY);

         HotX:=8;
         HotY:=7;
         MX:=MouseX;
         MY:=MouseY;

         {On mouse movement call handler }
         Define_EventHandler(1 {0000 0001} , @SVGA_Mouse_Fix);
      End;
End;

Procedure InitOnce;
Var i,X,Y: Byte;

Begin
   Writeln;
   Writeln('  Press <ENTER> to see a sketch of the function(s).');
   Write(  '  In the next screen, press <F1> to get a list of available keys.');
   Readln;
   GraphMode;

   MaxColors := 16{GetMaxColor};

   MaxX   :=GetMaxX+1;             { Min=0 ; Max=GetMaxX    }
                                   { Total GetMax+1 pixels  }
   MaxX_2 :=MaxX div 2;
   MaxX_10:=MaxX div 10;

   MaxY   :=GetMaxY+1;             { Min=0 ; Max=GetMaxY    }
                                   { Total GetMaxY+1 pixels }
   MaxY_2 :=MaxY div 2;
   MaxY_10:=MaxY div 10;

   T1:=0;
   T2:=0;
   For i:=1 to TF do
       Begin
          T1:=T1+Funcs[i]^.MaxX;
          T2:=T2+Funcs[i]^.MaxY;
       End;

   T1 := T1/TF;
   T2 := T2/TF;

   If T1<=0.5 then T1:=5;
   If T2<=0.5 then T2:=5;

   XM := T1 + (0.01*MaxX);
   YM := T2 + (0.01*MaxY);

   FuncColor  := Def_FuncColor;  { Color Defaults }
   FGColor    := Def_FGColor;
   BGColor    := Def_BGColor;
   AxisColor  := Def_AxisColor;

   MatchScale := False;          { Boolean Defaults }
   Asymptotes := True;
   DrawLines  := True;

   GridInfo   := All;            { Misc. Defaults }
   Axis       := Center;

   FuncStr    := 1;
   Graphs    := $FFFFFFFF; {all one(s) }
   FLoc[1].X := MaxX_2+5;
   FLoc[1].Y := 10;
   FLoc[2].X := 1;
   FLoc[2].Y := FLoc[1].Y;
   FLoc[3].X := FLoc[2].X;
   FLoc[3].Y := MaxY-20;
   FLoc[4].X := FLoc[1].X;
   FLoc[4].Y := FLoc[3].Y;

   Case Options.Mouse of
        0: MouseOn := MouseInstalled;
        1: MouseOn := True;
        2: MouseOn := False;
   End;

   BrackHeight := TextHeight('[');
   BrackWidth  := TextWidth('[');

   If MouseOn then
      Begin
         If Options.GDriver=SVGA256 then
            Begin
               For y:=0 to 15 do
                   For x:=0 to 15 do
                       Begin
                          { Screen Mask }
                          ColorCross[1,x,y] := $FF * GetBit(BWCross[y+1],15-x);

                          { Cursor Mask }
                          ColorCross[2,x,y] := $FF * GetBit(BWCross[y+17],15-x);
                       End;
            End;

         InitMouse(1,1);
      End;
End; {InitOnce}

{----------------------------------------------------------------------------}
Procedure Sketch_Function;

Var j,Tail: Integer;

Procedure Plot(Func: FuncPtr);
Var PutOnly : Boolean;
    V   : HelpType;
    XDot, YDot: Integer;
{    i: Integer;}
    Color: Byte;
    M,N,Y2: Float;
    YDot2: Integer;
    S: string;
    Temp: Complex;
    A,B,C: Boolean;

Procedure Draw_Function;
Var XDot: Integer;
Begin
   X:=-XM;
   PutOnly := True;
   For XDot:=0 to MaxX do
       Begin
         Y := Evaluate(Func,Func^.TypeOf,x) / ScaleY;

         If Is_Invalid_Operation then PutOnly :=True
         else
            {It seems that round doesn't accept numbers which not fit into REAL/LINT}
            If (MaxY_2-Y) <= -MaxY  then
               MoveTo(Xdot,-MaxY-10)
            else if (MaxY_2-Y) >= MaxY then
               MoveTo(XDot,MaxY+10)
            else
               Begin
                  YDot := Round(MaxY_2 - Y);
                  If (PutOnly) or (Not DrawLines) then
                     Begin
                        PutPixel(XDot, YDot, Color);
                        MoveTo(XDot,YDot);
                        PutOnly := False;
                     End
                  else
                     LineTo(XDot,YDot);
               End;

         X:=X+ScaleX;
       End;
End; {Draw_Function}

Procedure Draw_Asymptotes;
Var i: Integer;
Begin
   With Func^.Ratio do
      Begin
         SetLineStyle(DashedLn, 0, NormWidth);  {Dashed line }
         If GetBit(Func^.BInfo,2)=1 then { If there's an horiz. asym. }
            Begin
               Y := GetItem(Func^.MInfo,2) / ScaleY;
               If ( (MaxY_2-Y)>=-MaxY ) and ( (MaxY_2-Y)<=MaxY ) then
                  { If Y-Value is in-screen }
                  Begin
                     YDot := Round(MaxY_2 - Y);
                     Line(0, YDot, MaxX  , YDot);
                  End;
            End
         else If GetBit(Func^.BInfo,3)=1 then { If there's a curved asym. }
            Begin
               M := GetItem(Func^.MInfo,3);
               N := GetItem(Func^.MInfo,4);

               Y := (-M*XM+N) / ScaleY;
               Y2 := (M*XM+N) / ScaleY;

               YDot := Round(MaxY_2 - Y);
               YDot2 := Round(MaxY_2 - Y2);
               Line(0, YDot, MaxX  , YDot2);
            End;

         If Func^.TypeOf=Log then
            Begin
               XDot := Round(XM / ScaleX);
               Line(XDot ,0, XDot ,MaxY);  { Vertical Asym. }
            End
         else
         If (Func^.TypeOf=Rat) and
            (Denom.RootCount>0) and (Denom.NoRealRoots=False) then
            { If denominator got (real) zeros }
            Begin
               Find_Poly_Signs(Func^.Ratio.Denom, Func^.BInfo, V, Tail);
               For i:=1 to Tail-1 do
               Begin
                   { For every zero of the denominator }
                   CMake(V[i],0,Temp);

                   If (Not RootExist(Num,Temp)) and (V[i]+XM>=-MaxX) and (V[i]+XM<=MaxX) then
                       { If the same root doesn't exist in numerator
                         AND x-Value is in-screen }
                          Begin
                            XDot := Round((V[i]+XM) / ScaleX);
                            Line(XDot ,0, XDot ,MaxY);  { Vertical Asym. }
                          End;
               End; {For}
            End; {If}
         SetLineStyle(SolidLn, 0, NormWidth);  { Regular line }
      End;
End; {Draw_Asymptotes}

Begin
   Color := GetColor;
   If Asymptotes then Draw_Asymptotes;
   Draw_Function;
End; {Plot}

Function NextFuncColor(Var T: Byte): Byte;
Begin
   If T=MaxColors then T:=0
                  else Inc(T);

   If (T=BGColor) or (T=Black) then T:=NextFuncColor(T);

   NextFuncColor := T;
End; {CalcFuncColor}

Procedure CalcXYRes;
Var Mul: Comp;
    LogMul: Integer;
Begin
   Mul:=1;
   LogMul := 0;
   While Scalex*Mul<1 do
      Begin
         Mul := Mul*10;
         Inc(LogMul);
      End;
   XRes := LogMul;

   Mul:=1;
   LogMul := 0;
   While ScaleY*Mul<1 do
      Begin
         Mul := Mul*10;
         Inc(LogMul);
      End;
   YRes := LogMul;
End; { CalcXYRes }

Var S1,S2: String;
    XDot,YDot, YDot2, XAdd,YAdd: Integer;
    RelX, RelY: Integer;
    Color: Byte;
    Was: Boolean;
    Y2: Float;

type
  RGBRec = record
    RedVal, GreenVal, BlueVal : Integer;
     { Intensity values (values from 0 to 63) }
    Name: String;
    ColorNum: Integer;
     { The VGA color palette number as mapped into 16 color palette }
  end;
const
  { Table of suggested colors for VGA 16 color modes }
  Colors : array[0..15] of RGBRec = (
   ( RedVal:0;GreenVal:0;BlueVal:0;Name:'Black';ColorNum: 0),
   ( RedVal:0;GreenVal:0;BlueVal:40;Name:'Blue';ColorNum: 1),
   ( RedVal:0;GreenVal:40;BlueVal:0;Name:'Green';ColorNum: 2),
   ( RedVal:0;GreenVal:40;BlueVal:40;Name:'Cyan';ColorNum: 3),
   ( RedVal:40;GreenVal:7;BlueVal:7;Name:'Red';ColorNum: 4),
   ( RedVal:40;GreenVal:0;BlueVal:40;Name:'Magenta';ColorNum: 5),
   ( RedVal:40;GreenVal:30; BlueVal:0;Name:'Brown';ColorNum: 20),
   ( RedVal:49;GreenVal:49;BlueVal:49;Name:'Light Gray';ColorNum: 7),
   ( RedVal:26;GreenVal:26;BlueVal:26;Name:'Dark Gray';ColorNum: 56),
   ( RedVal:0;GreenVal:0;BlueVal:63;Name:'Light Blue';ColorNum: 57),
   ( RedVal:9;GreenVal:63;BlueVal:9;Name:'Light Green';ColorNum: 58),
   ( RedVal:0;GreenVal:63;BlueVal:63;Name:'Light Cyan';ColorNum: 59),
   ( RedVal:63;GreenVal:10;BlueVal:10;Name:'Light Red';ColorNum: 60),
   ( RedVal:44;GreenVal:0;BlueVal:63;Name:'Light Magenta';
          ColorNum: 61),
   ( RedVal:63;GreenVal:63;BlueVal:18;Name:'Yellow';ColorNum: 62),
   ( RedVal:63; GreenVal:63; BlueVal:63; Name: 'White'; ColorNum: 63)
  );

Procedure Draw_BackGround;
Var i,Tail: Integer;
Begin
   SetColor(FuncColor);
   If FuncStr>0 then
      Begin
         With FLoc[FuncStr] do MoveTo(X, Y);
         Was:=False;

         If GetBit(Graphs,0)=1 then
          For i:=1 to TF do
           If GetBit(Graphs,i)=1 then
             Begin
                If Was then OutText(';y= ')
                       else OutText('y= ');

                Was:=True;

                Case Funcs[i]^.TypeOf of
                  Pol: OutText(PolyStr(Funcs[i]^.Poly));

                  Rat:
                    Begin
                      YDot := TextHeight('y');
                      S1 := PolyStr(Funcs[i]^.Ratio.Num);
                      S2 := PolyStr(Funcs[i]^.Ratio.Denom);
                      j := BigInt(Length(S1),Length(S2));
                      S1:= PutBetween(0,j,S1);
                      S2:= PutBetween(0,j,S2);
                      OutTextXY(GetX,GetY-YDot,S1);
                      OutTextXY(GetX,GetY+YDot,S2);
                      For tail:=1 to j do OutText('-');
                    End;

                  Exp:
                    Begin
                      OutText(RealF(Funcs[i]^.BaseE)+'^x');
                      If Funcs[i]^.FreeE >0 then
                         OutText('+'+RealF(Funcs[i]^.FreeE))
                      else
                         If Funcs[i]^.FreeE <> 0 then
                            OutText(RealF(Funcs[i]^.FreeE));
                   End;

                  Log:
                    Begin
                      OutText('log X('+RealF(Funcs[i]^.BaseL)+')');
                      If Funcs[i]^.FreeL >0 then
                         OutText('+'+RealF(Funcs[i]^.FreeL))
                      else
                         If Funcs[i]^.FreeL <> 0 then
                            OutText(RealF(Funcs[i]^.FreeL));
                    End;

                End; {Case}
                SetColor(NextFuncColor(Color));
             End; {For i}
      End; {FuncStr>0}

   YAdd := Ord(Axis=Frame) * (MaxY_2-1);
   XAdd := Ord(Axis=Frame) * (MaxX_2-1);
   {XAdd & YAdd represent the middle pixels coordinates of the x & y axis }
   {That's of course true only when Ord(Axis=Frame)=1 >> Axis=Frame }

   If Axis<>Hide then
      Begin
         SetColor(AxisColor);
         Line(0, MaxY_2+YAdd, MaxX-1, MaxY_2+YAdd); { Draw X-Axis }
         Line(MaxX_2+XAdd, 0, MaxX_2+XAdd, MaxY-1); { Draw Y-Axis }

         If Axis=Center then
            Begin
               Line(MaxX-1,MaxY_2+YAdd, MaxX-6, MaxY_2+YAdd+5);{ and an arrow}
               Line(MaxX-1,MaxY_2+YAdd, MaxX-6, MaxY_2+YAdd-5);{ for X-Axis  }

               Line(MaxX_2+XAdd, 0, MaxX_2+XAdd+5, 5);       { Another one }
               Line(MaxX_2+XAdd, 0, MaxX_2+XAdd-5, 5);       { for Y-Axis  }

               SetColor(FGColor);

               OutTextXY( MaxX-10, MaxY_2+5, 'x');
               If FuncStr<>1 then OutTextXY( MaxX_2+5, 10 ,'y');
            End;
      End;
End; {Draw_Background}

Var i:Integer;
Begin
   If Options.GDriver=VGA then
      SetBKColor(BGColor)
   else
      With Colors[BGColor] do
           SetRGBPalette(0,RedVal,GreenVal,BlueVal{BGColor});

   ClearDevice;

   Color := FuncColor;
   Draw_Background; {Draws all background but grid info below }

   SetColor(FGColor);

   ScaleX:=XM / MaxX_2;
   If MatchScale then ScaleY := ScaleX
                 else ScaleY := YM / MaxY_2;

   If Options.GraphFP=-1 then CalcXYRes
      else Begin
              XRes := Options.GraphFP;
              YRes := Options.GraphFP;
           End;

   If (GridInfo<>None) and (Axis<>Hide) then
      Begin
         RelX := MaxX div 250;
         RelY := MaxY div 200;

         If YAdd>0 then Tail := -2*BrackHeight - 1*(RelY div 3)
                   else Tail :=  1*BrackHeight + 2*(RelY div 3);

         For XDot := 0 to 9 do
             Begin
                i := XDot*MaxX_10;
                Line(i , MaxY_2+YAdd+RelY, i , MaxY_2+YAdd-RelY);
                If GridInfo=All then
                   WriteXY(i, MaxY_2+YAdd+ Tail, XRes, (i-MaxX_2)*ScaleX );
             End;

         For YDot := 1 to 10 do
             Begin
                i := YDot*MaxY_10;
                If i >= MaxY then i:=MaxY-1; {1 pixel fix }

                If not ((Axis<>Frame) and (i=MaxY_2)) then
                { do not show two zeros in framed axis }
                   Begin
                      Line(MaxX_2+XAdd+RelX , i , MaxX_2+XAdd-RelX, i);
                      If (GridInfo=All) and not ((Axis=Frame) and (i=MaxY-1))
                      then
                         Begin
                            Str( (MaxY_2 - i)*ScaleY :0:Res2Axis(YRes), S1);
                            Tail := TextWidth(S1)+2*RelX;
                            WriteXY(MaxX_2+XAdd-Tail , i-BrackHeight , YRes, (MaxY_2 - i) * ScaleY);
                         End;
                   End;
             End; { For }
      End; {If }

   If Newton then
   Begin
{      Until Parse_Polynom(Func^.Poly,StrFunc)=True;}

      Y0 := Evaluate(Funcs[1],Funcs[1]^.TypeOf,X0);
      M := EvalPolynom(Funcs[1]^.PolyDiff,X0);
      N := Y0-M*X0;

      Y := (-M*XM+N) / ScaleY;
      Y2 := (M*XM+N) / ScaleY;

      YDot := Round(MaxY_2 - Y);
      YDot2 := Round(MaxY_2 - Y2);
      Line(0, YDot, MaxX  , YDot2);

      If M<>0 then X0 := X0 - Y0/M;
   End;

   Color := FuncColor;
   SetColor(FuncColor);

   If GetBit(Graphs,0)=1 then
      For i:=1 to TF do
          If GetBit(Graphs,i)=1 then
             Begin
                Plot(Funcs[i]);
                SetColor(NextFuncColor(Color));
             End;

End; {Sketch_Function}

Procedure Print_Screen;
Var Reg,kol,b,bbyt,b2: Word;
    Port: Byte;
Begin
   { MaxX should be equal to GetMaxX+1 }
   { MaxY should be equal to GetMaxY+1 }
   With Options do
        Port := LPTPort-1;

   Write(lst, chr(27), 'E');  { Reset Printer}
   Write(lst, chr(27), '*t',Int2Str(Options.PrintDPI),'R', chr(27), '*r0A');
   { '*t100R' means 100 DPI }
   { Set resolution to 100.. DPI , start Raster Graphics Mode}
   For reg := 0 to (Maxx-1) do
   begin
     Write(lst, chr(27), '*b', Maxy div 8, 'W');
     {Transfer raster graphics}
     For kol := (maxy div 8) - 1 downto 0 do
     begin
       bbyt := 0;
       For b := 0 to 7 do
       begin
         if getpixel(reg, kol * 8 + b) = 0 then
           b2 := 0
         else
           b2 := 1;
         bbyt := bbyt or (b2 shl b);
       end;
       b:=doInt17(chr(bbyt),Port);
     end;
   end;
   Write(lst, chr(27), '*rB');
   { End Raster Graphics mode - ???}
   Write(lst, chr(12));
   Write(lst, chr(27), 'E');
   { Reset Printer }
End; {Print_Screen}

Procedure ShowHelp;
Procedure HideCursor; Assembler;
Asm
  MOV   ax,$0100
  MOV   cx,$2607
  INT   $10
end;

Procedure ShowCursor; Assembler;
Asm
  MOV   ax,$0100
  MOV   cx,$0506
  INT   $10
end;
Var MouseX,MouseY: Word;
Begin
   GetMousePos(MouseX,MouseY); { Save old mouse position }
   RestoreCRTMode;
   TextBackground(Blue);
   TextColor(LightGray);
   ClrScr;
   TextColor(LightRed);
   Writeln(PutBetween(1,80,'Help'));
   TextColor(LightCyan);
   Writeln('          Scale :' );
   TextColor(LightGray);
   Writeln('      Switch scale - Auto / Match X-Y............................<S>');
   Writeln('      Increase X,Y scale respectively (Zoom Out).................<->');
   Writeln('      Decrease X,Y scale respectively (Zoom In)..................<+>');
   Writeln('      Restore default scale......................................<F9>');
   TextColor(LightCyan);
   Writeln('          Colors :');
   TextColor(LightGray);
   Writeln('      Graph......................................................<F3>');
   Writeln('      Foreground.................................................<F4>');
   Writeln('      Background.................................................<F5>');
   Writeln('      Axes.......................................................<F6>');
   Writeln('      Restore default colors.....................................<F10>');
   TextColor(LightCyan);
   Writeln('          Misc. :');
   TextColor(LightGray);
   Writeln('      Switch function : Show / Hide..............................<1..9>');
   Writeln('      Switch asym.    : Show / Hide..............................<A>');
   Writeln('      Print graph................................................<P>');
   Writeln('      Change mouse''s coordinates location : Up / Down / Hide.....<M>');
   Writeln('      Function''s name location : Up/Down/Right/Left/Hide.........<F>');
   Writeln('      Help.......................................................<F1>');
   Writeln('      Switch graph type: Lines / Dots............................<F2>');
   Writeln('      Change scale points : All / Lines Only / Hide..............<F7>');
   Writeln('      Change axes location : Center / Frame / Hide...............<F8>');
   Writeln('      Exit graph screen..........................................<ESC>/<ENTER>');
   Writeln;
   Write('Press <ENTER> to return to the previous screen');
   HideCursor;
   Readln;
   ShowCursor;
   GraphMode;
   If MouseOn then InitMouse(MouseX,MouseY);
End; {ShowHelp}

Const
  Enter = #13;
  {F1-F10 ASCII codes are #58 + #n}
  F1 = #59;
  F2 = #60;
  F3 = #61;
  F4 = #62;
  F5 = #63;
  F6 = #64;
  F7 = #65;
  F8 = #66;
  F9 = #67;
  F10 = #68;

Var FCSave, BGSave, GRSave, AXSave: Byte;
    Run: Boolean;
    OldX, OldY, MouseX, MouseY: Word;
    sx,sy: Integer;

Procedure ShowXY(X,Y: Word); Near;
Var StX, StY, S: String;
    XR, YR, T: Float;
    YDec: Integer;
Begin
   XR := ScaleX * (X-MaxX_2);
   YR := ScaleY * (MaxY_2-Y);
   With Options do
        Begin
           Str(XR:0:XRes,StX);
           Str(YR:0:YRes,StY);
        End;
   S := '['+StX+','+StY+']';
   YDec := Ord(MouseInfo=Top) * (MaxY-BrackHeight);
   OutTextXY(0,MaxY-BrackHeight-YDec,S);
End;

Var Ch: Char;
    T: Byte;
    i,j: Integer;
    mmx,mmy,SMask,CMask,NewPixel: Integer;
    But: Word;
    KeyFetch: Char;
    AnimateTimer: LongInt;

Begin
   OldExit := Nil;
   If Options.ErrorHandler then
      Begin
         OldExit := ExitProc;

         SetJump(MyAddr);     {We'll get back here whenever a run time    }
                              {error occurs                               }
                              {This line should always be before          }
                              {ExitProc:=@MyExit;                         }
         ExitProc:=@MyExit;
      End;

   InitOnce;
   Quit:=False;
   Run:=True;
   X0:=0;
   Repeat
      If Run then
         Begin
            Sketch_Function;

            If MouseOn then
               Begin
                  If Options.GDriver=SVGA256 then
                  Begin
                     GetMousePos(mx,my);
                     mmx := mx-HotX;
                     mmy := my-HotY;

                     For j:=0 to 15 do
                         For i:=0 to 15 do
                             OldPixels[i,j]:=getpixel(mmX+i,mmY+j);

                     For j:=0 to 15 do
                         For i:=0 to 15 do
                         Begin
                            SMask := ColorCross[1,i,j];
                            CMask := ColorCross[2,i,j];
                            NewPixel := (White and CMask and SMask) XOR OldPixels[i,j];
                            PutPixel(mmX+i,mmY+j,NewPixel);
                         End;
                  End;

                  ShowMouse;
{                  Conditional_Hidden(0,MaxY-1,100,MaxY);}
               End;
         End
           else
             Run:=True;

      Newton:=False;
      If (MouseOn) and (MouseInfo<>Nothing) then
      Begin
         { Make sure there will be a first run in the while statement down}
         { (unless a key is pressed) }
         GetMousePos(OldX,OldY);
         OldX:=OldX+1;
         OldY:=OldY+1;

         KeyFetch:=#0;
         AnimateTimer:=0;

         While (Not KeyPressed) and (KeyFetch=#0) and not
               (Newton and (AnimateTimer>=50000)) do
         Begin
            If Newton then
               AnimateTimer := AnimateTimer + 1;

            {Need to install an event handler instead ...}
            GetMousePos(MouseX,MouseY);
            GetMouseBut(But);

            If But>0 then
               Case But of
                    (1*1)+(0*2)+(0*4): KeyFetch:='+';  {Only bit 0 is set}
                    (0*1)+(1*2)+(0*4): KeyFetch:='-';  {         1       }
               End;

            If (OldX<>MouseX) or (OldY<>MouseY) then
               Begin
                  SetColor(BGColor);
                  ShowXY(OldX,OldY);
                  SetColor(FGColor);
                  ShowXY(MouseX,MouseY);
                  OldX := MouseX;
                  OldY := MouseY;
               End;
         End;
      End;

      If AnimateTimer>=50000 then KeyFetch:='N';

      If KeyFetch<>#0 then Ch:=KeyFetch
                      else Ch:= UpCase(ReadKey);

      Case Ch of
           #0:  { Function key (F1..F10,Shift,Alt etc.) }
             Begin
                Case ReadKey of
                     F1: ShowHelp;
                     F2: DrawLines  := Not DrawLines;
                     {Seems like black is having trouble
                      to be seen on a non-black background.
                      So Palette Color 0 will be reserved
                      (by us or by someone else ?) for use as the
                      background color.
                     }
                     F3: If FuncColor=MaxColors then FuncColor:=1
                                                else Inc(FuncColor);

                     F4: If FGColor=MaxColors then FGColor:=1
                                              else Inc(FGColor);

                     F5: If BGColor=MaxColors then BGColor:=1
                                              else Inc(BGColor);

                     F6: If AxisColor=MaxColors then AxisColor:=1
                                                else Inc(AxisColor);

                     F7: If GridInfo=None then GridInfo:=All
                                          else Inc(GridInfo);

                     F8:
                      Begin
                         If Axis=Hide then Axis:=Center
                                      else Inc(Axis);

                         If (Axis=Frame) and (MouseInfo=Buttom) then
                            MouseInfo:=Top;
                      End;

                     F9:
                       Begin
                          XM := T1 + (0.01*MaxX);
                          YM := T2 + (0.01*MaxY);
                       End;
                     F10:
                       Begin
                          FuncColor:= Def_FuncColor;
                          FGColor  := Def_FGColor;
                          BGColor  := Def_BGColor;
                          AxisColor:= Def_AxisColor;
                       End;
                else
                  Run:=False;
                End;
             End;
      '0'..'9':
             Begin
                T := Ord(Ch)-Ord('0');
                If GetBit(Graphs,T)=1 then
                   Graphs := ClrBit(Graphs,T)
                else
                   Graphs := SetBit(Graphs,T);
             End;
           'A': Asymptotes := Not Asymptotes;
           'B': X0:=10;

           'S': MatchScale := Not MatchScale;
           'P':   {Print It !}
              If Printer_Ok then
                 Begin
                    { Modify colors so the page won't be Black }
                    FCSave := FuncColor;
                    BGSave := BGColor;
                    GRSave := FGColor;
                    AxSave := AxisColor;
                    If BGColor <> Black then BGColor := Black;
                    If FuncColor = Black then FuncColor := Cyan;
                    If FGColor = Black then FGColor := Yellow;
                    If AxisColor = Black then AxisColor := White;
                    Sketch_Function;
                    Print_Screen;
                    FuncColor := FCSave;
                    BGColor := BGSave;
                    FGColor := GRSave;
                    AxisColor := AxSave;
                    If MouseOn then HideMouse;
                    Sketch_Function; { Draw with new colors }
                 End
              else
                 Run:=False;

           'M': If MouseInfo=Nothing then MouseInfo:=Buttom
                                     else Inc(MouseInfo);

           'N': {Magic} ;

           'F': If FuncStr=4 then FuncStr := 0
                             else Inc(FuncStr);

           '-':
              Begin
                 XM := 1.1*XM; {10% Increase}
                 YM := 1.1*YM;
              End;
           '+':
              Begin
                 XM := 0.9*XM; {10% Decrease}
                 YM := 0.9*YM;
              End;

   Esc,Enter: Quit:=True;

      else
        Run:=False;
      End;

   Until Quit=True;

   { Disable Defined Mouse Handler }
   If Options.GDriver=SVGA256 then Define_EventHandler(0, Nil);

   CloseGraph;
   NewInit:=True;
   If (Options.ErrorHandler) and (Assigned(OldExit)) then
      ExitProc := OldExit; {Release error handler}
End; {Sketch}

{----------------------------------------------------------------------------}
Procedure DisposeOf(Var Funcs: FuncArr);

{$IFDEF DEBUG}
Procedure ShowMemDiff;
Begin
   Writeln('Start: ',FMem);
   Writeln('End  : ',MemAvail);
   Writeln('Diff : ',MemAvail-FMem);
   Writeln('Press <ENTER> to continue.');
   Readln;
End;
{$ENDIF}

Procedure DisposeMInfo(Var MInfo: MInfoPtr);
Var Idx, Next: MInfoPtr;
Begin
   Idx := MInfo;
   While Idx<>Nil do
         Begin
            Next := Idx^.Link;
            Dispose(Idx);
            Idx := Next;
         End;
End; {DisposeMInfo}

Var i: Byte;
Begin
   For i:=TF downto 1 do { For every function allocated }
       Begin
          With Funcs[i]^ do
          Begin
             DisposeMInfo(MInfo);
             Case TypeOf of
                  Pol:
                     Begin
                        DisposePoly(Poly);
                        DisposePoly(PolyDiff);
                     End;

                  Rat:
                     Begin
                        DisposePoly(Ratio.Num);
                        DisposePoly(Ratio.Denom);
                        DisposePoly(RatioDiff.Num);
                        DisposePoly(RatioDiff.Denom);
                     End;
             End; {Case}
          End;
          Dispose(Funcs[i]); { Dispose pointer's memory }
   End;

   {$IFDEF DEBUG}
   ShowMemDiff;
   {$ENDIF}

End; {DisposeOf}

{----------------------------------------------------------------------------}
Function DoYouWantAnother: Boolean;
{ Sketch another function ? }
Var Key: Char;

Begin
   ClrScr;
   Writeln;
   Writeln;
   Write('     Do you want to analyze another function (Y/N) ? ');
   Repeat
     Key:=UpCase(ReadKey);
   Until Key in ['Y','N'];

   DoYouWantAnother := Key='Y';
End; {DoYouWantAnother}

{----------------------------------------------------------------------------}
Function AddAnotherFunc: Boolean;
{ Add another function to sketch ? }
Var Key: Char;

Begin
   If TF=MaxFuncs then
      Begin
         AddAnotherFunc:=False;
         Exit;
      End;

   ClrScr;
   Writeln;
   Writeln;
   Write('     Do you want to add another function (Y/N) ? ');
   Repeat
     Key:=UpCase(ReadKey);
   Until Key in ['Y','N'];

   AddAnotherFunc := Key='Y';
End; {AddAnotherFunc}

{----------------------------------------------------------------------------}
Procedure Quit;
{ Back to original text mode }
Begin
   TextMode(LastMode);
End; {Quit}

{----------------------------------------------------------------------------}
Procedure InitFunc(Var Func: FuncPtr);
Begin
   New(Func);   { Memory allocation }

   FillChar(Func^,SizeOf(Func^),0);  { Clean garbage - Needed !? }

   { Init. of pointers, counters, etc. }
   With Func^ do
        Begin
           BInfo := 0;
           MInfo := Nil;
           Case TypeOf of
                Pol:
                   Begin
                      Poly.RootCount:=0;
                      PolyDiff.RootCount:=0;
                      Poly.Roots := Nil;
                      Poly.NoRealRoots := True;
                      PolyDiff.Roots := Nil;
                      PolyDiff.NoRealRoots := True;
                   End;

                Rat: With Ratio do
                   Begin
                      Num.RootCount:=0;
                      Num.Roots := Nil;
                      Num.NoRealRoots := True;

                      Denom.RootCount:=0;
                      Denom.Roots := Nil;
                      Denom.NoRealRoots := True;

                      RatioDiff.Num.RootCount:=0;
                      RatioDiff.Num.Roots := Nil;
                      RatioDiff.Num.NoRealRoots := True;

                      RatioDiff.Denom.RootCount:=0;
                      RatioDiff.Denom.Roots := Nil;
                      RatioDiff.Denom.NoRealRoots := True;
                   End;
             { Exp, Log, General : No init needed }

           End; {Case}
        End;
End; {InitFunc}

{----------------------------------------------------------------------------}
End. {Function_Analyzer_Unit}
