Unit MSMouse;

{ Copyright (C) 1999 Idan Nof }
{ Read the attached files "README.TXT" and/or "GPL.TXT" for more details }

{ This unit implements standard calls to a Microsoft Compatible
  Mouse Driver
}

Interface

Const
   MouseInstalled: Boolean = False;

Function  ResetMouse: Word;
{ Function 0 - Init Mouse
  Returns -1 is sucessfull, otherwise 0
}

Procedure ShowMouse;
{ Function 1 - Show Mouse Cursor}

Procedure HideMouse;
{ Function 2 - Hide Mouse Cursor }

Procedure GetMousePos(Var {Buttons,}X,Y: Word);
{ Function 3 - Get Mouse Position & Buttons Status }

Procedure GetMouseBut(Var Buttons: Word);
{ Function 3 - Get Mouse Position & Buttons Status }

Procedure SetMousePos(X,Y: Word);
{ Function 4 - Set Mouse Position }

Procedure SetRangeX(MinX,MaxX: Word);
{ Function 7 - Set Mouse Min/Max horizontal range }

Procedure SetRangeY(MinY,MaxY: Word);
{ Function 8 - Set Mouse Min/Max vertical range }

Procedure Set_GraphMouseCursor(Var Cursor; HotSpotX, HotSpotY: Word);
{ Function 9 - Define Graphics cursor & its hotspot}

Procedure Define_EventHandler(Mask: Word; Handler: Pointer);
{ Function 12 - Define an event handler }

Procedure SetMickey(Horiz, Vert: Word);
{ Function 15 - Set Motion/Pixel ratio }

Procedure Conditional_Hidden(x1,y1,x2,y2: Word);
{ Function 16 - Conditional Hidden Cursor }

Procedure Swap_EventHandlers(Var Mask: Word; Var Handler: Pointer);

Function SetScale1to1: Word;

Function SetScale2to1: Word;

Function SetResolution(Horiz,Vert: Word): Word;

Implementation

Const
    MIO = $33;   {Mouse Software Interrupt}
    OK  = $FFFF; { = -1 }

Function ResetMouse: Word; Assembler;
ASM
  MOV  AX,0
  INT  MIO
End;

Procedure ShowMouse; Assembler;
ASM
  MOV  AX,1
  INT  MIO
END;

Procedure HideMouse; Assembler;
ASM
  MOV  AX,2
  INT  MIO
END;

Procedure Set_GraphMouseCursor(Var Cursor; HotSpotX, HotSpotY: Word); Assembler;
ASM
  MOV  AX,9
  MOV  BX,HotSpotX
  MOV  CX,HotSpotY
  LES  DX, Cursor   { Make [ES:DX] pointer to cursor }
  INT  MIO
END;

Procedure SetMousePos(X,Y: Word); Assembler;
ASM
  MOV  AX,4
  MOV  CX,X
  MOV  DX,Y
  INT  MIO
END;

Procedure GetMousePos(Var {Buttons,}X,Y: Word); Assembler;
{ BX holds button status
  CX, DX Holds X & Y Pointer location, resprectively
}

ASM
  MOV  AX,3
  INT  MIO

  { Store Buttons State }
{  LES  SI, Buttons
  MOV  [ES:SI], BX}

  { Store XY State }
  LES  SI, X
  MOV  [ES:SI], CX
  LES  SI, Y
  MOV  [ES:SI], DX
END;

Procedure GetMouseBut(Var Buttons: Word); Assembler;
ASM
  MOV  AX,3
  INT  MIO

  { Store Buttons State }
  LES  SI, Buttons
  MOV  [ES:SI], BX
End;

Procedure SetRangeX(MinX,MaxX: Word); Assembler;
ASM
  MOV  AX,7
  MOV  CX,MinX
  MOV  DX,MaxX
  INT  MIO
END;

Procedure SetRangeY(MinY,MaxY: Word); Assembler;
ASM
   MOV  AX,8
   MOV  CX,MinY
   MOV  DX,MaxY
   INT  MIO
END;

Procedure Define_EventHandler(Mask: Word; Handler: Pointer); Assembler;
ASM
  MOV  AX,12
  MOV  CX,Mask
  LES  DX,Handler
  INT  MIO
End;

Procedure SetMickey(Horiz, Vert: Word); Assembler;
ASM
  MOV  AX,15
  MOV  CX,Horiz
  MOV  DX,Vert
  INT  MIO
END;

Procedure Conditional_Hidden(x1,y1,x2,y2: Word); Assembler;
ASM
  MOV  AX,16
  MOV  CX,x1
  MOV  DX,y1
  MOV  SI,x2
  MOV  DI,y2
  INT  MIO
END;

Procedure Swap_EventHandlers(Var Mask: Word; Var Handler: Pointer); Assembler;
{ Untested !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! }
ASM
  MOV  AX,20

  { Load Mask into CX }
  LES  BX,Mask
  MOV  CX,[ES:BX]

  { Load Handler Address }
  LES  DX,Handler
  INT  MIO

  { Save Old Handler Address }
  PUSH DX
  PUSH ES

  LES  BX,Handler   { Load 'Handler' address }
  POP  AX           { Pop Segment }
  MOV  [ES:BX],AX   { Store into 'Handler' }
  POP  AX           { Pop Offset }
  MOV  [ES:BX+2],AX { Store into 'Handler' }

  LES  BX,Mask
  MOV  [ES:BX], CX
END;

Function SetScale1to1: Word; Assembler;
ASM
  MOV  AX, $A5
  MOV  BX, $E6
  INT  MIO
END;

Function SetScale2to1: Word; Assembler;
ASM
  MOV  AX, $A5
  MOV  BX, $E7
  INT  MIO
END;

Function SetResolution(Horiz,Vert: Word): Word; Assembler;
ASM
  MOV  AX, $1A
  MOV  BX, Horiz
  MOV  CX, Vert
  INT  MIO
END;

Begin
   MouseInstalled := ( ResetMouse=OK );
End.